import re
from collections import namedtuple

import numpy as np
import scipy.sparse as sp

from Orange.data import DiscreteVariable, Domain
from Orange.data.sql.table import SqlTable
from Orange.preprocess.util import _RefuseDataInConstructor
from Orange.statistics import distribution, contingency, util as ut
from Orange.statistics.basic_stats import BasicStats
from Orange.util import Reprable
from .transformation import Transformation
from . import _discretize

__all__ = ["EqualFreq", "EqualWidth", "EntropyMDL", "DomainDiscretizer",
           "decimal_binnings", "get_bins"]


class Discretizer(Transformation):
    """Value transformer that returns an index of the bin for the given value.
    """
    def __init__(self, variable, points):
        super().__init__(variable)
        self.points = points

    @staticmethod
    def digitize(x, bins):
        if sp.issparse(x):
            if len(bins):
                x.data = np.digitize(x.data, bins)
            else:
                x = sp.csr_matrix(x.shape)
            return x
        else:
            return np.digitize(x, bins) if len(bins) else [0]*len(x)

    def transform(self, c):
        if sp.issparse(c):
            return self.digitize(c, self.points)
        elif c.size:
            return np.where(np.isnan(c), np.NaN, self.digitize(c, self.points))
        else:
            return np.array([], dtype=int)

    @staticmethod
    def _fmt_interval(low, high, formatter):
        assert low is not None or high is not None
        assert low is None or high is None or low < high
        if low is None or np.isinf(low):
            return f"< {formatter(high)}"
        if high is None or np.isinf(high):
            return f"≥ {formatter(low)}"
        return f"{formatter(low)} - {formatter(high)}"

    @classmethod
    def create_discretized_var(cls, var, points):
        def fmt(val):
            sval = var.str_val(val)
            # For decimal numbers, remove trailing 0's and . if no decimals left
            if re.match(r"^\d+\.\d+", sval):
                return sval.rstrip("0").rstrip(".")
            return sval

        lpoints = list(points)
        if lpoints:
            values = [
                cls._fmt_interval(low, high, fmt)
                for low, high in zip([-np.inf] + lpoints, lpoints + [np.inf])]
            to_sql = BinSql(var, lpoints)
        else:
            values = ["single_value"]
            to_sql = SingleValueSql(values[0])

        dvar = DiscreteVariable(name=var.name, values=values,
                                compute_value=cls(var, points),
                                sparse=var.sparse)
        dvar.source_variable = var
        dvar.to_sql = to_sql
        return dvar


class BinSql:
    def __init__(self, var, points):
        self.var = var
        self.points = points

    def __call__(self):
        return 'width_bucket(%s, ARRAY%s::double precision[])' % (
            self.var.to_sql(), str(self.points))


class SingleValueSql:
    def __init__(self, value):
        self.value = value

    def __call__(self):
        return "'%s'" % self.value


class Discretization(Reprable):
    """Abstract base class for discretization classes."""
    def __call__(self, data, variable):
        """
        Compute discretization of the given variable on the given data.
        Return a new variable with the appropriate domain
        (:obj:`Orange.data.DiscreteVariable.values`) and transformer
        (:obj:`Orange.data.Variable.compute_value`).
        """
        raise NotImplementedError(
            "Subclasses of 'Discretization' need to implement "
            "the call operator")


class EqualFreq(Discretization):
    """Discretization into bins with approximately equal number of data
    instances.

    .. attribute:: n

        Number of bins (default: 4). The actual number may be lower if the
        variable has less than n distinct values.
    """
    def __init__(self, n=4):
        self.n = n

    # noinspection PyProtectedMember
    def __call__(self, data, attribute):
        if type(data) == SqlTable:
            att = attribute.to_sql()
            quantiles = [(i + 1) / self.n for i in range(self.n - 1)]
            query = data._sql_query(
                ['quantile(%s, ARRAY%s)' % (att, str(quantiles))],
                use_time_sample=1000)
            with data._execute_sql_query(query) as cur:
                points = sorted(set(cur.fetchone()[0]))
        else:
            d = distribution.get_distribution(data, attribute)
            points = _discretize.split_eq_freq(d, self.n)
        return Discretizer.create_discretized_var(
            data.domain[attribute], points)

class EqualWidth(Discretization):
    """Discretization into a fixed number of bins with equal widths.

    .. attribute:: n

        Number of bins (default: 4).
    """
    def __init__(self, n=4):
        self.n = n

    # noinspection PyProtectedMember
    def __call__(self, data, attribute, fixed=None):
        if fixed:
            min, max = fixed[attribute.name]
            points = self._split_eq_width(min, max)
        else:
            if type(data) == SqlTable:
                stats = BasicStats(data, attribute)
                points = self._split_eq_width(stats.min, stats.max)
            else:
                values = data[:, attribute]
                values = values.X if values.X.size else values.Y
                min, max = ut.nanmin(values), ut.nanmax(values)
                points = self._split_eq_width(min, max)
        return Discretizer.create_discretized_var(
            data.domain[attribute], points)

    def _split_eq_width(self, min, max):
        if np.isnan(min) or np.isnan(max) or min == max:
            return []
        dif = (max - min) / self.n
        return [min + (i + 1) * dif for i in range(self.n - 1)]


BinDefinition = namedtuple("BinDefinition", ("start", "nbins", "width"))


def decimal_binnings(
        data, *, min_width=0, min_bins=2, max_bins=50,
        min_unique=5, add_unique=None,
        factors=(20, 10, 5, 2, 1, 0.5, 0.25, 0.2, 0.1, 0.05, 0.025, 0.02, 0.01),
        return_defs=False):
    """
    Find a set of nice splits of data into bins

    The function first computes the scaling factor that is, the power of 10
    that brings the interval of values within [0, 1]. For instances, if the
    numbers come from interaval 10004001 and 10007005, the width of the
    interval is 3004, so the scaling factor is 1000.

    The function next considers bin widths that are products of scaling and
    different factors from 20 to 0.01 that make sense in decimal scale
    (see default value for argument `factors`). For each width, it rounds the
    minimal value down to this width and the maximal value up, and it computes
    the number of bins of that width that fit between these two values.
    If the number of bins is between `min_bins` and `max_bins`, and the width
    is at least `min_width`, this is a valid interval.

    If the data has no more than `min_unique` unique values, the function will
    add a set of bins that put each value into its own bin.

    If the data has no more than `add_unique` values, that last bins will put
    each value into its own bin.

    Args:
        data (np.ndarray):
            vector of data points; values may repeat, and nans and infs are
            filtered out.
        min_width (float): minimal bin width
        min_bins (int): minimal number of bins
        max_bins (int):
            maximal number of bins; the number of bins will never exceed the
            number of unique values
        min_unique (int):
            if the number of unique values are less or equal to `min_unique`,
            the function returns a single binning that matches that values in
            the data
        add_unique (int):
            similar to `min_unique` except that such bins are added to the list
        factors (list of float):
            The factors with which the scaling is multiplied. Default is
            `(20, 10, 5, 2, 1, 0.5, 0.25, 0.2, 0.1, 0.05, 0.025, 0.02, 0.01)`,
            so if scaling is 1000, considered bin widths are 20000, 10000,
            5000, 2000, 1000, 500, 250, 200, 100, 50, 25, 20 and 10.
        return_defs (bool): If set to `True`, the function returns a list of
            instances of `BinDefinition`, otherwise a list of bin boundaries.

    Returns:
        bin_boundaries (list of np.ndarray): a list of bin boundaries,
            including the top boundary of the last interval, hence the list
            size equals the number bins + 1. These array match the `bin`
            argument of `numpy.histogram`.

            This is returned if `return_defs` is left `True`.

        bin_definition (list of BinDefinition):
            `BinDefinition` is a named tuple containing the beginning of the
            first bin (`start`), number of bins (`nbins`) and their widths
            (`width`). The last value can also be a `nd.array` with `nbins + 1`
            elements, which describes bins of unequal width and is used for
            binnings that match the unique values in the data (see `min_unique`
            and `add_unique`).

            This is returned if `return_defs` is `False`.
    """
    def unique_bins():
        if len(unique) >= 2:
            # make the last bin the same width as the one before
            last_boundary = 2 * unique[-1] - unique[-2]
        else:
            last_boundary = unique[0] + 1
        return BinDefinition(
            unique[0], len(unique), np.hstack((unique, [last_boundary])))

    unique = np.unique(data)
    unique = unique[np.isfinite(unique)]
    if not unique.size:
        raise ValueError("no valid (non-nan) data")
    mn, mx = unique[0], unique[-1]
    if mn == mx or len(unique) <= min_unique:
        bins = unique_bins()
        if not return_defs:
            bins = get_bins(bins)
        return [bins]

    diff = mx - mn
    f10 = 10 ** -np.floor(np.log10(diff))
    bins = []
    max_bins = min(max_bins, len(unique))
    for f in factors:
        width = f / f10
        if width < min_width:
            continue
        mn_ = np.floor(mn / width) * width
        mx_ = np.ceil(mx / width) * width
        nbins = np.round((mx_ - mn_) / width)
        if min_bins <= nbins <= max_bins \
                and (not bins or bins[-1].nbins != nbins):
            bins.append(BinDefinition(mn_, nbins, width))

    if add_unique is not None and len(unique) <= add_unique:
        if bins and bins[-1].nbins == len(unique):
            del bins[-1]
        bins.append(unique_bins())
        if len(unique) < min_unique:
            del bins[:-1]

    if not return_defs:
        bins = [get_bins(bin) for bin in bins]

    return bins


def get_bins(bin_def: BinDefinition):
    """
    Return a `np.ndarray` corresponding to interval
    Args:
        bin_def (BinDefinition):
            definition of bins. a named tuple containing the beginning of the
            first bin (`start`), number of bins (`nbins`) and their widths
            (`width`). The last value can also be a `nd.array` with `nbins + 1`
            elements, in which case the function returns this as a result.

    Returns:
        bin boundaries (np.ndarray):
        bin boundaries including the top boundary of the last interval, hence
        the list size equals `bin_def.nbins + 1`. This array matches the
        `bin` argument of `numpy.histogram`.
    """
    if isinstance(bin_def.width, np.ndarray):
        return bin_def.width
    return bin_def.start + bin_def.width * np.arange(bin_def.nbins + 1)


# noinspection PyPep8Naming
class EntropyMDL(Discretization):
    """
    Discretization into bins inferred by recursively splitting the values to
    minimize the class-entropy. The procedure stops when further splits would
    decrease the entropy for less than the corresponding increase of minimal
    description length (MDL). [FayyadIrani93].

    If there are no suitable cut-off points, the procedure returns a single bin,
    which means that the new feature is constant and can be removed.

    .. attribute:: force

        Induce at least one cut-off point, even when its information
        gain is lower than MDL (default: False).

    """
    def __init__(self, force=False):
        self.force = force

    def __call__(self, data, attribute):
        cont = contingency.get_contingency(data, attribute)
        values, I = cont.values, cont.counts.T
        cut_ind = np.array(self._entropy_discretize_sorted(I, self.force))
        if len(cut_ind) > 0:
            # "the midpoint between each successive pair of examples" (FI p.1)
            points = (values[cut_ind] + values[cut_ind - 1]) / 2.
        else:
            points = []
        return Discretizer.create_discretized_var(
            data.domain[attribute], points)

    @classmethod
    def _normalize(cls, X, axis=None, out=None):
        """
        Normalize `X` array so it sums to 1.0 over the `axis`.

        Parameters
        ----------
        X : array
            Array to normalize.
        axis : optional int
            Axis over which the resulting array sums to 1.
        out : optional array
            Output array of the same shape as X.
        """
        X = np.asarray(X, dtype=float)
        scale = np.sum(X, axis=axis, keepdims=True)
        if out is None:
            return X / scale
        else:
            if out is not X:
                assert out.shape == X.shape
                out[:] = X
            out /= scale
            return out

    @classmethod
    def _entropy_normalized(cls, D, axis=None):
        """
        Compute the entropy of distribution array `D`.

        `D` must be a distribution (i.e. sum to 1.0 over `axis`)

        Parameters
        ----------
        D : array
            Distribution.
        axis : optional int
            Axis of `D` along which to compute the entropy.

        """
        # req: (np.sum(D, axis=axis) >= 0).all()
        # req: (np.sum(D, axis=axis) <= 1).all()
        # req: np.all(np.abs(np.sum(D, axis=axis) - 1) < 1e-9)

        D = np.asarray(D)
        Dc = np.clip(D, np.finfo(D.dtype).eps, 1.0)
        return - np.sum(D * np.log2(Dc), axis=axis)

    @classmethod
    def _entropy(cls, D, axis=None):
        """
        Compute the entropy of distribution `D`.

        Parameters
        ----------
        D : array
            Distribution.
        axis : optional int
            Axis of `D` along which to compute the entropy.

        """
        D = cls._normalize(D, axis=axis)
        return cls._entropy_normalized(D, axis=axis)

    @classmethod
    def _entropy1(cls, D):
        """
        Compute the entropy of distributions in `D`
        (one per each row).
        """
        D = cls._normalize(D)
        return _discretize.entropy_normalized1(D)

    @classmethod
    def _entropy2(cls, D):
        """
        Compute the entropy of distributions in `D`
        (one per each row).
        """
        D = cls._normalize(D, axis=1)
        return _discretize.entropy_normalized2(D)

    @classmethod
    def _entropy_cuts_sorted(cls, CS):
        """
        Return the class information entropy induced by partitioning
        the `CS` distribution at all N-1 candidate cut points.

        Parameters
        ----------
        CS : (N, K) array of class distributions.
        """
        CS = np.asarray(CS)
        # |--|-------|--------|
        #  S1    ^       S2
        # S1 contains all points which are <= to cut point
        # Cumulative distributions for S1 and S2 (left right set)
        # i.e. a cut at index i separates the CS into S1Dist[i] and S2Dist[i]
        S1Dist = np.cumsum(CS, axis=0)[:-1]
        S2Dist = np.cumsum(CS[::-1], axis=0)[-2::-1]

        # Entropy of S1[i] and S2[i] sets
        ES1 = cls._entropy2(S1Dist)
        ES2 = cls._entropy2(S2Dist)

        # Number of cases in S1[i] and S2[i] sets
        S1_count = np.sum(S1Dist, axis=1)
        S2_count = np.sum(S2Dist, axis=1)

        # Number of all cases
        S_count = np.sum(CS)

        ES1w = ES1 * S1_count / S_count
        ES2w = ES2 * S2_count / S_count

        # E(A, T; S) Class information entropy of the partition S
        E = ES1w + ES2w

        return E, ES1, ES2

    @classmethod
    def _entropy_discretize_sorted(cls, C, force=False):
        """
        Entropy discretization on a sorted C.

        :param C: (N, K) array of class distributions.

        """
        E, ES1, ES2 = cls._entropy_cuts_sorted(C)
        # TODO: Also get the left right distribution counts from
        # entropy_cuts_sorted,

        # Note the + 1
        if len(E) == 0:
            return []
        cut_index = np.argmin(E) + 1

        # Distribution of classed in S1, S2 and S
        S1_c = np.sum(C[:cut_index], axis=0)
        S2_c = np.sum(C[cut_index:], axis=0)
        S_c = S1_c + S2_c

        ES = cls._entropy1(np.sum(C, axis=0))
        ES1, ES2 = ES1[cut_index - 1], ES2[cut_index - 1]

        # Information gain of the best split
        Gain = ES - E[cut_index - 1]
        # Number of different classes in S, S1 and S2
        k = float(np.sum(S_c > 0))
        k1 = float(np.sum(S1_c > 0))
        k2 = float(np.sum(S2_c > 0))

        assert k > 0
        delta = np.log2(3 ** k - 2) - (k * ES - k1 * ES1 - k2 * ES2)
        N = float(np.sum(S_c))

        if Gain > np.log2(N - 1) / N + delta / N:
            # Accept the cut point and recursively split the subsets.
            left, right = [], []
            if k1 > 1 and cut_index > 1:
                left = cls._entropy_discretize_sorted(C[:cut_index, :])
            if k2 > 1 and cut_index < len(C) - 1:
                right = cls._entropy_discretize_sorted(C[cut_index:, :])
            return left + [cut_index] + [i + cut_index for i in right]
        elif force:
            return [cut_index]
        else:
            return []


class DomainDiscretizer(_RefuseDataInConstructor, Reprable):
    """Discretizes all continuous features in the data.

    .. attribute:: method

        Feature discretization method (instance of
        :obj:`Orange.preprocess.Discretization`). If `None` (default),
        :class:`Orange.preprocess.EqualFreq` with 4 intervals is
        used.

    .. attribute:: clean

        If `True`, features discretized into a single interval constant are
        removed. This is useful for discretization methods that infer the
        number of intervals from the data, such as
        :class:`Orange.preprocess.EntropyMDL` (default: `True`).

    .. attribute:: discretize_class

        Determines whether a target is also discretized if it is continuous.
        (default: `False`)
    """
    def __init__(self, discretize_class=False, method=None, clean=True,
                 fixed=None):
        self.discretize_class = discretize_class
        self.method = method
        self.clean = clean
        self.fixed = fixed

    def __call__(self, data, fixed=None):
        """
        Compute and return discretized domain.

        :param data: Data to discretize.
        """

        def transform_list(s, fixed=None):
            new_vars = []
            for var in s:
                if var.is_continuous:
                    if fixed and var.name in fixed.keys():
                        nv = method(data, var, fixed)
                    else:
                        nv = method(data, var)
                    if not self.clean or len(nv.values) > 1:
                        new_vars.append(nv)
                else:
                    new_vars.append(var)
            return new_vars
        if self.method is None:
            method = EqualFreq(n=4)
        else:
            method = self.method
        domain = data.domain
        new_attrs = transform_list(domain.attributes, fixed or self.fixed)
        if self.discretize_class:
            new_classes = transform_list(domain.class_vars)
        else:
            new_classes = domain.class_vars
        return Domain(new_attrs, new_classes)



# Test methods with long descriptive names can omit docstrings
# pylint: disable=missing-docstring

import unittest
from unittest import mock

from Orange.data import Table
from Orange.widgets.data.owsql import OWSql
from Orange.widgets.tests.base import WidgetTest
from Orange.tests.sql.base import DataBaseTest as dbt


class TestOWSqlConnected(WidgetTest, dbt):
    def setUpDB(self):
        # pylint: disable=attribute-defined-outside-init
        self.widget = self.create_widget(OWSql)
        self.params, _ = self.create_iris_sql_table()
        self.iris = Table("iris")

    def tearDownDB(self):
        self.drop_iris_sql_table()

    @dbt.run_on(["postgres"])
    def test_connection(self):
        """Test if a connection to the database can be established"""
        self.set_connection_params()
        self.widget.connect()

        self.assertFalse(self.widget.Error.connection.is_shown())
        self.assertIsNotNone(self.widget.database_desc)
        tables = ["Select a table", "Custom SQL"]
        self.assertTrue(set(self.widget.tables).issuperset(set(tables)))

    @dbt.run_on(["postgres"])
    def test_output_iris(self):
        """Test if iris data can be fetched from database"""
        self.assertIsNone(self.get_output(self.widget.Outputs.data))

        self.set_connection_params()
        self.widget.connect()

        idx = list(map(str, self.widget.tables)).index("iris")
        self.widget.tablecombo.setCurrentIndex(idx)
        self.widget.select_table()

        output = self.get_output(self.widget.Outputs.data)
        self.assertIsNotNone(output)
        self.assertEqual(len(output), len(self.iris))
        iris_domain = set(map(str, self.iris.domain.attributes))
        output_domain = set(map(str, output.domain.attributes))
        self.assertTrue(output_domain.issuperset(iris_domain))

    def set_connection_params(self):
        """Set database connection parameters on widget"""
        port = ''
        if self.params['port'] is not None:
            port += ':' + str(self.params['port'])
        self.widget.servertext.setText(self.params['host'] + port)
        self.widget.databasetext.setText(self.params['database'])
        self.widget.usernametext.setText(self.params['user'])
        self.widget.passwordtext.setText(self.params['password'])


class TestOWSql(WidgetTest):

    @mock.patch('Orange.widgets.data.owsql.Backend')
    def test_missing_extension(self, mock_backends):
        """Test for correctly handled missing backend extension"""
        backend = mock.Mock()
        backend().display_name = "PostgreSQL"
        backend().missing_extension = ["missing extension"]
        backend().list_tables.return_value = []
        mock_backends.available_backends.return_value = [backend]

        widget = self.create_widget(OWSql)

        self.assertTrue(widget.Warning.missing_extension.is_shown())
        self.assertTrue(widget.download)
        self.assertFalse(widget.downloadcb.isEnabled())

    @mock.patch('Orange.widgets.data.owsql.Backend')
    def test_non_postgres(self, mock_backends):
        """Test if download is enforced for non postgres backends"""
        backend = mock.Mock()
        backend().display_name = "database"
        del backend().missing_extension
        backend().list_tables.return_value = []
        mock_backends.available_backends.return_value = [backend]

        widget = self.create_widget(OWSql)

        self.assertTrue(widget.download)
        self.assertFalse(widget.downloadcb.isEnabled())

    @mock.patch('Orange.widgets.data.owsql.Table')
    @mock.patch('Orange.widgets.data.owsql.SqlTable')
    @mock.patch('Orange.widgets.data.owsql.Backend')
    def test_restore_table(self, mock_backends, mock_sqltable, mock_table):
        """Test if selected table is restored from settings"""
        backend = mock.Mock()
        backend().display_name = "database"
        del backend().missing_extension
        backend().list_tables.return_value = ["a", "b", "c"]
        mock_backends.available_backends.return_value = [backend]
        mock_sqltable().approx_len.return_value = 100

        settings = {'__version__': 2,
                    'host': '',
                    'port': '',
                    'guess_values': False,
                    'download': False,
                    'table': 'b'}

        widget = self.create_widget(OWSql, stored_settings=settings)
        self.assertEqual(widget.tablecombo.currentText(), "b")


if __name__ == "__main__":
    unittest.main()



from contextlib import contextmanager
import os
import pickle
from unittest.mock import Mock, patch

import numpy as np
import scipy.sparse as sp

from AnyQt.QtCore import QRectF, QPointF
from AnyQt.QtTest import QSignalSpy
from AnyQt.QtWidgets import (
    QComboBox, QSpinBox, QDoubleSpinBox, QSlider
)

from orangewidget.tests.base import (
    # pylint: disable=unused-import
    GuiTest, WidgetTest as WidgetTestBase, DummySignalManager, DEFAULT_TIMEOUT
)
from Orange.base import SklModel, Model
from Orange.classification.base_classification import (
    LearnerClassification, ModelClassification
)
from Orange.data import (
    Table, Domain, DiscreteVariable, ContinuousVariable, Variable
)
from Orange.modelling import Fitter
from Orange.preprocess import RemoveNaNColumns, Randomize, Continuize
from Orange.preprocess.preprocess import PreprocessorList
from Orange.regression.base_regression import (
    LearnerRegression, ModelRegression
)
from Orange.widgets.tests.utils import simulate
from Orange.widgets.utils.annotated_data import (
    ANNOTATED_DATA_FEATURE_NAME, ANNOTATED_DATA_SIGNAL_NAME
)
from Orange.widgets.utils.owlearnerwidget import OWBaseLearner
from Orange.widgets.widget import OWWidget


class WidgetTest(WidgetTestBase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        Variable._clear_all_caches()


class TestWidgetTest(WidgetTest):
    """Meta tests for widget test helpers"""

    def test_process_events_handles_timeouts(self):
        with self.assertRaises(TimeoutError):
            self.process_events(until=lambda: False, timeout=0)

    def test_minimum_size(self):
        return  # skip this test


class BaseParameterMapping:
    """Base class for mapping between gui components and learner's parameters
    when testing learner widgets.

    Parameters
    ----------
    name : str
        Name of learner's parameter.

    gui_element : QWidget
        Gui component who's corresponding parameter is to be tested.

    values: list
        List of values to be tested.

    getter: function
        It gets component's value.

    setter: function
        It sets component's value.
    """

    def __init__(self, name, gui_element, values, getter, setter,
                 problem_type="both"):
        self.name = name
        self.gui_element = gui_element
        self.values = values
        self.get_value = getter
        self.set_value = setter
        self.problem_type = problem_type

    def __str__(self):
        if self.problem_type == "both":
            return self.name
        else:
            return "%s (%s)" % (self.name, self.problem_type)


class DefaultParameterMapping(BaseParameterMapping):
    """Class for mapping between gui components and learner's parameters
    when testing unchecked properties and therefore default parameters
    should be used.

    Parameters
    ----------
    name : str
        Name of learner's parameter.

    default_value: str, int,
        Value that should be used by default.
    """

    def __init__(self, name, default_value):
        super().__init__(name, None, [default_value],
                         lambda: default_value, lambda x: None)


class ParameterMapping(BaseParameterMapping):
    """Class for mapping between gui components and learner parameters
    when testing learner widgets

    Parameters
    ----------
    name : str
        Name of learner's parameter.

    gui_element : QWidget
        Gui component who's corresponding parameter is to be tested.

    values: list, mandatory for ComboBox, optional otherwise
        List of values to be tested. When None, it is set according to
        component's type.

    getter: function, optional
        It gets component's value. When None, it is set according to
        component's type.

    setter: function, optional
        It sets component's value. When None, it is set according to
        component's type.
    """

    def __init__(self, name, gui_element, values=None,
                 getter=None, setter=None, **kwargs):
        super().__init__(
            name, gui_element,
            values or self._default_values(gui_element),
            getter or self._default_get_value(gui_element, values),
            setter or self._default_set_value(gui_element, values),
            **kwargs)

    @staticmethod
    def get_gui_element(widget, attribute):
        return widget.controlled_attributes[attribute][0].control

    @classmethod
    def from_attribute(cls, widget, attribute, parameter=None):
        return cls(parameter or attribute, cls.get_gui_element(widget, attribute))

    @staticmethod
    def _default_values(gui_element):
        if isinstance(gui_element, (QSpinBox, QDoubleSpinBox, QSlider)):
            return [gui_element.minimum(), gui_element.maximum()]
        else:
            raise TypeError("{} is not supported".format(gui_element))

    @staticmethod
    def _default_get_value(gui_element, values):
        if isinstance(gui_element, (QSpinBox, QDoubleSpinBox, QSlider)):
            return lambda: gui_element.value()
        elif isinstance(gui_element, QComboBox):
            return lambda: values[gui_element.currentIndex()]
        else:
            raise TypeError("{} is not supported".format(gui_element))

    @staticmethod
    def _default_set_value(gui_element, values):
        if isinstance(gui_element, (QSpinBox, QDoubleSpinBox, QSlider)):
            return lambda val: gui_element.setValue(val)
        elif isinstance(gui_element, QComboBox):
            def fun(val):
                value = values.index(val)
                gui_element.activated.emit(value)
                gui_element.setCurrentIndex(value)

            return fun
        else:
            raise TypeError("{} is not supported".format(gui_element))


class WidgetLearnerTestMixin:
    """Base class for widget learner tests.

    Contains init method to set up testing parameters and test methods.

    All widget learner tests should extend it (beside extending WidgetTest
    class as well). Learners with extra parameters, which can be set on the
    widget, should override self.parameters list in the setUp method. The
    list should contain mapping: learner parameter - gui component.
    """

    widget = None  # type: OWBaseLearner

    def init(self):
        cls_ds = Table(datasets.path("testing_dataset_cls"))
        reg_ds = Table(datasets.path("testing_dataset_reg"))

        if issubclass(self.widget.LEARNER, Fitter):
            self.data = cls_ds
            self.valid_datasets = (cls_ds, reg_ds)
            self.inadequate_dataset = ()
            self.learner_class = Fitter
            self.model_class = Model
            self.model_name = 'Model'
        elif issubclass(self.widget.LEARNER, LearnerClassification):
            self.data = cls_ds
            self.valid_datasets = (cls_ds,)
            self.inadequate_dataset = (reg_ds,)
            self.learner_class = LearnerClassification
            self.model_class = ModelClassification
            self.model_name = 'Classifier'
        else:
            self.data = reg_ds
            self.valid_datasets = (reg_ds,)
            self.inadequate_dataset = (cls_ds,)
            self.learner_class = LearnerRegression
            self.model_class = ModelRegression
            self.model_name = 'Predictor'

        self.parameters = []

    def test_has_unconditional_apply(self):
        self.assertTrue(hasattr(self.widget, "unconditional_apply"))

    def test_input_data(self):
        """Check widget's data with data on the input"""
        self.assertEqual(self.widget.data, None)
        self.send_signal("Data", self.data)
        self.assertEqual(self.widget.data, self.data)
        self.wait_until_stop_blocking()

    def test_input_data_disconnect(self):
        """Check widget's data and model after disconnecting data from input"""
        self.send_signal("Data", self.data)
        self.assertEqual(self.widget.data, self.data)
        self.widget.apply_button.button.click()
        self.wait_until_stop_blocking()
        self.send_signal("Data", None)
        self.wait_until_stop_blocking()
        self.assertEqual(self.widget.data, None)
        self.assertIsNone(self.get_output(self.widget.Outputs.model))

    def test_input_data_learner_adequacy(self):
        """Check if error message is shown with inadequate data on input"""
        for inadequate in self.inadequate_dataset:
            self.send_signal("Data", inadequate)
            self.widget.apply_button.button.click()
            self.wait_until_stop_blocking()
            self.assertTrue(self.widget.Error.data_error.is_shown())
        for valid in self.valid_datasets:
            self.send_signal("Data", valid)
            self.wait_until_stop_blocking()
            self.assertFalse(self.widget.Error.data_error.is_shown())

    def test_input_preprocessor(self):
        """Check learner's preprocessors with an extra pp on input"""
        randomize = Randomize()
        self.send_signal("Preprocessor", randomize)
        self.assertEqual(
            randomize, self.widget.preprocessors,
            'Preprocessor not added to widget preprocessors')
        self.widget.apply_button.button.click()
        self.wait_until_stop_blocking()
        self.assertEqual(
            (randomize,), self.widget.learner.preprocessors,
            'Preprocessors were not passed to the learner')

    def test_input_preprocessors(self):
        """Check multiple preprocessors on input"""
        pp_list = PreprocessorList([Randomize(), RemoveNaNColumns()])
        self.send_signal("Preprocessor", pp_list)
        self.widget.apply_button.button.click()
        self.wait_until_stop_blocking()
        self.assertEqual(
            (pp_list,), self.widget.learner.preprocessors,
            '`PreprocessorList` was not added to preprocessors')

    def test_input_preprocessor_disconnect(self):
        """Check learner's preprocessors after disconnecting pp from input"""
        randomize = Randomize()
        self.send_signal("Preprocessor", randomize)
        self.widget.apply_button.button.click()
        self.wait_until_stop_blocking()
        self.assertEqual(randomize, self.widget.preprocessors)

        self.send_signal("Preprocessor", None)
        self.widget.apply_button.button.click()
        self.wait_until_stop_blocking()
        self.assertIsNone(self.widget.preprocessors,
                          'Preprocessors not removed on disconnect.')

    def test_output_learner(self):
        """Check if learner is on output after apply"""
        initial = self.get_output("Learner")
        self.assertIsNotNone(initial, "Does not initialize the learner output")
        self.widget.apply_button.button.click()
        newlearner = self.get_output("Learner")
        self.assertIsNot(initial, newlearner,
                         "Does not send a new learner instance on `Apply`.")
        self.assertIsNotNone(newlearner)
        self.assertIsInstance(newlearner, self.widget.LEARNER)

    def test_output_model(self):
        """Check if model is on output after sending data and apply"""
        self.assertIsNone(self.get_output(self.widget.Outputs.model))
        self.widget.apply_button.button.click()
        self.assertIsNone(self.get_output(self.widget.Outputs.model))
        self.send_signal('Data', self.data)
        self.widget.apply_button.button.click()
        self.wait_until_stop_blocking()
        model = self.get_output(self.widget.Outputs.model)
        self.assertIsNotNone(model)
        self.assertIsInstance(model, self.widget.LEARNER.__returns__)
        self.assertIsInstance(model, self.model_class)

    def test_output_learner_name(self):
        """Check if learner's name properly changes"""
        new_name = "Learner Name"
        self.widget.apply_button.button.click()
        self.assertEqual(self.widget.learner.name,
                         self.widget.name_line_edit.text()
                         or self.widget.name_line_edit.placeholderText())
        self.widget.name_line_edit.setText(new_name)
        self.widget.apply_button.button.click()
        self.wait_until_stop_blocking()
        self.assertEqual(self.get_output("Learner").name, new_name)

    def test_output_model_name(self):
        """Check if model's name properly changes"""
        new_name = "Model Name"
        self.widget.name_line_edit.setText(new_name)
        self.send_signal("Data", self.data)
        self.widget.apply_button.button.click()
        self.wait_until_stop_blocking()
        self.assertEqual(self.get_output(self.widget.Outputs.model).name, new_name)

    def test_output_model_picklable(self):
        """Check if model can be pickled"""
        self.send_signal("Data", self.data)
        self.widget.apply_button.button.click()
        self.wait_until_stop_blocking()
        model = self.get_output(self.widget.Outputs.model)
        pickle.dumps(model)

    def _get_param_value(self, learner, param):
        if isinstance(learner, Fitter):
            # Both is just a was to indicate to the tests, fitters don't
            # actually support this
            if param.problem_type == "both":
                problem_type = learner.CLASSIFICATION
            else:
                problem_type = param.problem_type
            return learner.get_params(problem_type).get(param.name)
        else:
            return learner.params.get(param.name)

    def test_parameters_default(self):
        """Check if learner's parameters are set to default (widget's) values
        """
        for dataset in self.valid_datasets:
            self.send_signal("Data", dataset)
            self.widget.apply_button.button.click()
            self.wait_until_stop_blocking()
            for parameter in self.parameters:
                # Skip if the param isn't used for the given data type
                if self._should_check_parameter(parameter, dataset):
                    self.assertEqual(
                        self._get_param_value(self.widget.learner, parameter),
                        parameter.get_value())

    def test_parameters(self):
        """Check learner and model for various values of all parameters"""
        # Test params on every valid dataset, since some attributes may apply
        # to only certain problem types
        for dataset in self.valid_datasets:
            self.send_signal("Data", dataset)
            self.wait_until_stop_blocking()

            for parameter in self.parameters:
                # Skip if the param isn't used for the given data type
                if not self._should_check_parameter(parameter, dataset):
                    continue

                assert isinstance(parameter, BaseParameterMapping)

                for value in parameter.values:
                    parameter.set_value(value)
                    self.widget.apply_button.button.click()
                    self.wait_until_stop_blocking()
                    param = self._get_param_value(self.widget.learner, parameter)
                    self.assertEqual(
                        param, parameter.get_value(),
                        "Mismatching setting for parameter '%s'" % parameter)
                    self.assertEqual(
                        param, value,
                        "Mismatching setting for parameter '%s'" % parameter)
                    param = self._get_param_value(self.get_output("Learner"), parameter)
                    self.assertEqual(
                        param, value,
                        "Mismatching setting for parameter '%s'" % parameter)

                    if issubclass(self.widget.LEARNER, SklModel):
                        model = self.get_output(self.widget.Outputs.model)
                        if model is not None:
                            self.assertEqual(self._get_param_value(model, parameter), value)
                            self.assertFalse(self.widget.Error.active)
                        else:
                            self.assertTrue(self.widget.Error.active)

    def test_params_trigger_settings_changed(self):
        """Check that the learner gets updated whenever a param is changed."""
        for dataset in self.valid_datasets:
            self.send_signal("Data", dataset)
            self.wait_until_stop_blocking()

            for parameter in self.parameters:
                # Skip if the param isn't used for the given data type
                if not self._should_check_parameter(parameter, dataset):
                    continue

                assert isinstance(parameter, BaseParameterMapping)
                # Set the mock here so we can include the param name in the
                # error message, so if any test fails, we see where
                # We mock `apply` and not `settings_changed` since that's
                # sometimes connected with Qt signals, which are not directly
                # called
                self.widget.apply = Mock(name="apply(%s)" % parameter)
                # Since the settings only get updated when the value actually
                # changes, find a value that isn't the same as the current
                # value and try with that
                new_value = [x for x in parameter.values
                             if x != parameter.get_value()][0]
                parameter.set_value(new_value)
                self.widget.apply.assert_called_once_with()

    @staticmethod
    def _should_check_parameter(parameter, data):
        """Should the param be passed into the learner given the data"""
        return ((parameter.problem_type == "classification" and
                 data.domain.has_discrete_class) or
                (parameter.problem_type == "regression" and
                 data.domain.has_continuous_class) or
                (parameter.problem_type == "both"))


class WidgetOutputsTestMixin:
    """Class for widget's outputs testing.

    Contains init method to set up testing parameters and a test method, which
    checks Selected Data and (Annotated) Data outputs.

    Since widgets have different ways of selecting data instances, _select_data
    method should be implemented when subclassed. The method should assign
    value to selected_indices parameter.

    If output's expected domain differs from input's domain, parameter
    same_input_output_domain should be set to False.

    If Selected Data and Data domains differ, override method
    _compare_selected_annotated_domains.
    """

    def init(self):
        Variable._clear_all_caches()
        self.data = Table("iris")
        self.same_input_output_domain = True

    def test_outputs(self, timeout=DEFAULT_TIMEOUT):
        self.send_signal(self.signal_name, self.signal_data)

        if self.widget.isBlocking():
            spy = QSignalSpy(self.widget.blockingStateChanged)
            self.assertTrue(spy.wait(timeout))

        # check selected data output
        self.assertIsNone(self.get_output("Selected Data"))

        # check annotated data output
        feature_name = ANNOTATED_DATA_FEATURE_NAME
        annotated = self.get_output(ANNOTATED_DATA_SIGNAL_NAME)
        self.assertEqual(0, np.sum([i[feature_name] for i in annotated]))

        # select data instances
        selected_indices = self._select_data()

        # check selected data output
        selected = self.get_output("Selected Data")
        n_sel, n_attr = len(selected), len(self.data.domain.attributes)
        self.assertGreater(n_sel, 0)
        self.assertEqual(selected.domain == self.data.domain,
                         self.same_input_output_domain)
        np.testing.assert_array_equal(selected.X[:, :n_attr],
                                      self.data.X[selected_indices])
        self.assertEqual(selected.attributes, self.data.attributes)

        # check annotated data output
        annotated = self.get_output(ANNOTATED_DATA_SIGNAL_NAME)
        self.assertEqual(n_sel, np.sum([i[feature_name] for i in annotated]))
        self.assertEqual(annotated.attributes, self.data.attributes)

        # compare selected and annotated data domains
        self._compare_selected_annotated_domains(selected, annotated)

        # check output when data is removed
        self.send_signal(self.signal_name, None)
        self.assertIsNone(self.get_output("Selected Data"))
        self.assertIsNone(self.get_output(ANNOTATED_DATA_SIGNAL_NAME))

    def _select_data(self):
        raise NotImplementedError("Subclasses should implement select_data")

    def _compare_selected_annotated_domains(self, selected, annotated):
        selected_vars = selected.domain.variables + selected.domain.metas
        annotated_vars = annotated.domain.variables + annotated.domain.metas
        self.assertLess(set(selected_vars), set(annotated_vars))


class ProjectionWidgetTestMixin:
    """Class for projection widget testing"""

    def init(self):
        Variable._clear_all_caches()
        self.data = Table("iris")

    def _select_data(self):
        rect = QRectF(QPointF(-20, -20), QPointF(20, 20))
        self.widget.graph.select_by_rectangle(rect)
        return self.widget.graph.get_selection()

    def _compare_selected_annotated_domains(self, selected, annotated):
        selected_vars = selected.domain.variables
        annotated_vars = annotated.domain.variables
        self.assertLessEqual(set(selected_vars), set(annotated_vars))

    def test_setup_graph(self, timeout=DEFAULT_TIMEOUT):
        """Plot should exist after data has been sent in order to be
        properly set/updated"""
        self.send_signal(self.widget.Inputs.data, self.data)

        if self.widget.isBlocking():
            spy = QSignalSpy(self.widget.blockingStateChanged)
            self.assertTrue(spy.wait(timeout))

        self.assertIsNotNone(self.widget.graph.scatterplot_item)

    def test_default_attrs(self, timeout=DEFAULT_TIMEOUT):
        """Check default values for 'Color', 'Shape', 'Size' and 'Label'"""
        self.send_signal(self.widget.Inputs.data, self.data)
        self.assertIs(self.widget.attr_color, self.data.domain.class_var)
        self.assertIsNone(self.widget.attr_label)
        self.assertIsNone(self.widget.attr_shape)
        self.assertIsNone(self.widget.attr_size)
        if self.widget.isBlocking():
            spy = QSignalSpy(self.widget.blockingStateChanged)
            self.assertTrue(spy.wait(timeout))
        self.send_signal(self.widget.Inputs.data, None)
        self.assertIsNone(self.widget.attr_color)

    def test_attr_models(self):
        """Check possible values for 'Color', 'Shape', 'Size' and 'Label'"""
        self.send_signal(self.widget.Inputs.data, self.data)
        controls = self.widget.controls
        self.assertEqual(len(controls.attr_color.model()), 8)
        self.assertEqual(len(controls.attr_shape.model()), 3)
        self.assertTrue(5 < len(controls.attr_size.model()) < 8)
        self.assertEqual(len(controls.attr_label.model()), 8)

        # color and label should contain all variables
        # size should contain only continuous variables
        # shape should contain only discrete variables
        for var in self.data.domain.variables + self.data.domain.metas:
            self.assertIn(var, controls.attr_color.model())
            self.assertIn(var, controls.attr_label.model())
            if var.is_continuous:
                self.assertIn(var, controls.attr_size.model())
                self.assertNotIn(var, controls.attr_shape.model())
            if var.is_discrete:
                self.assertNotIn(var, controls.attr_size.model())
                self.assertIn(var, controls.attr_shape.model())

    def test_attr_label_metas(self, timeout=DEFAULT_TIMEOUT):
        """Set 'Label' from string meta attribute"""
        cont = Continuize(multinomial_treatment=Continuize.AsOrdinal)
        data = cont(Table("zoo"))
        self.send_signal(self.widget.Inputs.data, data)
        if self.widget.isBlocking():
            spy = QSignalSpy(self.widget.blockingStateChanged)
            self.assertTrue(spy.wait(timeout))
        simulate.combobox_activate_item(self.widget.controls.attr_label,
                                        data.domain[-1].name)

    def test_handle_primitive_metas(self):
        """Set 'Color' from continuous meta attribute"""
        d, attrs = self.data.domain, self.data.domain.attributes
        data = self.data.transform(Domain(attrs[:2], d.class_vars, attrs[2:]))
        self.send_signal(self.widget.Inputs.data, data)
        simulate.combobox_activate_item(self.widget.controls.attr_color,
                                        data.domain.metas[0].name)

    def test_datasets(self, timeout=DEFAULT_TIMEOUT):
        """Test widget for datasets with missing values and constant features"""
        for ds in datasets.datasets():
            self.send_signal(self.widget.Inputs.data, ds)
            if self.widget.isBlocking():
                spy = QSignalSpy(self.widget.blockingStateChanged)
                self.assertTrue(spy.wait(timeout))

    def test_none_data(self):
        """Test widget for empty dataset"""
        self.send_signal(self.widget.Inputs.data, self.data[:0])

    def test_plot_once(self, timeout=DEFAULT_TIMEOUT):
        """Test if data is plotted only once but committed on every input change"""
        table = Table("heart_disease")
        self.widget.setup_plot = Mock()
        self.widget.commit = Mock()
        self.send_signal(self.widget.Inputs.data, table)
        self.widget.setup_plot.assert_called_once()
        self.widget.commit.assert_called_once()

        if self.widget.isBlocking():
            spy = QSignalSpy(self.widget.blockingStateChanged)
            self.assertTrue(spy.wait(timeout))
        self.widget.setup_plot.assert_called_once()
        self.widget.commit.assert_called_once()

        self.widget.commit.reset_mock()
        self.send_signal(self.widget.Inputs.data_subset, table[::10])
        self.widget.setup_plot.assert_called_once()
        self.widget.commit.assert_called_once()

    def test_subset_data_color(self, timeout=DEFAULT_TIMEOUT):
        self.send_signal(self.widget.Inputs.data, self.data)

        if self.widget.isBlocking():
            spy = QSignalSpy(self.widget.blockingStateChanged)
            self.assertTrue(spy.wait(timeout))

        self.send_signal(self.widget.Inputs.data_subset, self.data[:10])
        subset = [brush.color().name() == "#46befa" for brush in
                  self.widget.graph.scatterplot_item.data['brush'][:10]]
        other = [brush.color().name() == "#000000" for brush in
                 self.widget.graph.scatterplot_item.data['brush'][10:]]
        self.assertTrue(all(subset))
        self.assertTrue(all(other))

    def test_class_density(self, timeout=DEFAULT_TIMEOUT):
        """Check class density update"""
        self.send_signal(self.widget.Inputs.data, self.data)
        self.widget.cb_class_density.click()
        if self.widget.isBlocking():
            spy = QSignalSpy(self.widget.blockingStateChanged)
            self.assertTrue(spy.wait(timeout))
        self.send_signal(self.widget.Inputs.data, None)
        self.widget.cb_class_density.click()

    def test_dragging_tooltip(self):
        """Dragging tooltip depends on data being jittered"""
        text = self.widget.graph.tiptexts[0]
        self.send_signal(self.widget.Inputs.data, Table("heart_disease"))
        self.assertEqual(self.widget.graph.tip_textitem.toPlainText(), text)
        self.widget.graph.controls.jitter_size.setValue(1)
        self.assertGreater(self.widget.graph.tip_textitem.toPlainText(), text)

    def test_sparse_data(self, timeout=DEFAULT_TIMEOUT):
        """Test widget for sparse data"""
        table = Table("iris").to_sparse()
        self.assertTrue(sp.issparse(table.X))
        self.send_signal(self.widget.Inputs.data, table)
        if self.widget.isBlocking():
            spy = QSignalSpy(self.widget.blockingStateChanged)
            self.assertTrue(spy.wait(timeout))
        self.send_signal(self.widget.Inputs.data_subset, table[::30])
        self.assertEqual(len(self.widget.subset_data), 5)

    def test_invalidated_embedding(self, timeout=DEFAULT_TIMEOUT):
        """Check if graph has been replotted when sending same data"""
        self.widget.graph.update_coordinates = Mock()
        self.widget.graph.update_point_props = Mock()
        self.send_signal(self.widget.Inputs.data, self.data)
        if self.widget.isBlocking():
            spy = QSignalSpy(self.widget.blockingStateChanged)
            self.assertTrue(spy.wait(timeout))

        self.widget.graph.update_coordinates.assert_called()
        self.widget.graph.update_point_props.assert_called()

        self.widget.graph.update_coordinates.reset_mock()
        self.widget.graph.update_point_props.reset_mock()
        self.send_signal(self.widget.Inputs.data, self.data)
        if self.widget.isBlocking():
            spy = QSignalSpy(self.widget.blockingStateChanged)
            self.assertTrue(spy.wait(timeout))

        self.widget.graph.update_coordinates.assert_not_called()
        self.widget.graph.update_point_props.assert_called_once()

    def test_saved_selection(self, timeout=DEFAULT_TIMEOUT):
        self.send_signal(self.widget.Inputs.data, self.data)
        if self.widget.isBlocking():
            spy = QSignalSpy(self.widget.blockingStateChanged)
            self.assertTrue(spy.wait(timeout))

        self.widget.graph.select_by_indices(list(range(0, len(self.data), 10)))
        settings = self.widget.settingsHandler.pack_data(self.widget)
        w = self.create_widget(self.widget.__class__, stored_settings=settings)

        self.send_signal(self.widget.Inputs.data, self.data, widget=w)
        if w.isBlocking():
            spy = QSignalSpy(w.blockingStateChanged)
            self.assertTrue(spy.wait(timeout))

        self.assertEqual(np.sum(w.graph.selection), 15)
        np.testing.assert_equal(self.widget.graph.selection, w.graph.selection)

    def test_send_report(self, timeout=DEFAULT_TIMEOUT):
        """Test report """
        self.send_signal(self.widget.Inputs.data, self.data)
        self.widget.report_button.click()
        if self.widget.isBlocking():
            spy = QSignalSpy(self.widget.blockingStateChanged)
            self.assertTrue(spy.wait(timeout))
        self.send_signal(self.widget.Inputs.data, None)
        self.widget.report_button.click()

    def test_hidden_effective_variables(self, timeout=DEFAULT_TIMEOUT):
        hidden_var1 = ContinuousVariable("c1")
        hidden_var1.attributes["hidden"] = True
        hidden_var2 = ContinuousVariable("c2")
        hidden_var2.attributes["hidden"] = True
        class_vars = [DiscreteVariable("cls", values=["a", "b"])]
        table = Table(Domain([hidden_var1, hidden_var2], class_vars),
                      np.array([[0., 1.], [2., 3.]]),
                      np.array([[0.], [1.]]))
        self.send_signal(self.widget.Inputs.data, table)
        if self.widget.isBlocking():
            spy = QSignalSpy(self.widget.blockingStateChanged)
            self.assertTrue(spy.wait(timeout))
        self.send_signal(self.widget.Inputs.data, table)


class AnchorProjectionWidgetTestMixin(ProjectionWidgetTestMixin):
    def test_embedding_missing_values(self):
        table = Table("heart_disease")
        table.X[0] = np.nan
        self.send_signal(self.widget.Inputs.data, table)
        self.assertFalse(np.all(self.widget.valid_data))
        output = self.get_output(ANNOTATED_DATA_SIGNAL_NAME)
        embedding_mask = np.all(np.isnan(output.metas[:, :2]), axis=1)
        np.testing.assert_array_equal(~embedding_mask, self.widget.valid_data)
        # reload
        self.send_signal(self.widget.Inputs.data, table)

    def test_sparse_data(self):
        table = Table("iris")
        table.X = sp.csr_matrix(table.X)
        self.assertTrue(sp.issparse(table.X))
        self.send_signal(self.widget.Inputs.data, table)
        self.assertTrue(self.widget.Error.sparse_data.is_shown())
        self.send_signal(self.widget.Inputs.data_subset, table[::30])
        self.assertEqual(len(self.widget.subset_data), 5)
        self.send_signal(self.widget.Inputs.data, None)
        self.assertFalse(self.widget.Error.sparse_data.is_shown())

    def test_manual_move(self):
        data = self.data.copy()
        data[1, 0] = np.nan
        nvalid, nsample = len(self.data) - 1, self.widget.SAMPLE_SIZE
        self.send_signal(self.widget.Inputs.data, data)
        self.widget.graph.select_by_indices(list(range(0, len(data), 10)))

        # remember state
        selection = self.widget.graph.selection.copy()

        # simulate manual move
        self.widget._manual_move_start()
        self.widget._manual_move(0, 1, 1)
        self.assertEqual(len(self.widget.graph.scatterplot_item.data), nsample)
        self.widget._manual_move_finish(0, 1, 2)

        # check new state
        self.assertEqual(len(self.widget.graph.scatterplot_item.data), nvalid)
        np.testing.assert_equal(self.widget.graph.selection, selection)


class datasets:
    @staticmethod
    def path(filename):
        dirname = os.path.join(os.path.dirname(__file__), "datasets")
        return os.path.join(dirname, filename)

    @classmethod
    def missing_data_1(cls):
        """
        Dataset with 3 continuous features (X{1,2,3}) where all the columns
        and rows contain at least one NaN value.

        One discrete class D with NaN values
        Mixed continuous/discrete/string metas ({X,D,S}M)

        Returns
        -------
        data : Orange.data.Table
        """
        return Table(cls.path("missing_data_1.tab"))

    @classmethod
    def missing_data_2(cls):
        """
        Dataset with 3 continuous features (X{1,2,3}) where all the columns
        and rows contain at least one NaN value and X1, X2 are constant.

        One discrete constant class D with NaN values.
        Mixed continuous/discrete/string class metas ({X,D,S}M)

        Returns
        -------
        data : Orange.data.Table
        """
        return Table(cls.path("missing_data_2.tab"))

    @classmethod
    def missing_data_3(cls):
        """
        Dataset with 3 discrete features D{1,2,3} where all the columns and
        rows contain at least one NaN value

        One discrete class D with NaN values
        Mixes continuous/discrete/string metas ({X,D,S}M)

        Returns
        -------
        data : Orange.data.Table
        """
        return Table(cls.path("missing_data_3.tab"))

    @classmethod
    def data_one_column_vals(cls, value=np.nan):
        """
        Dataset with two continuous features and one discrete. One continuous
        columns has custom set values (default nan).

        Returns
        -------
        data : Orange.data.Table
        """
        table = Table(
            Domain(
                [ContinuousVariable("a"),
                 ContinuousVariable("b"),
                 DiscreteVariable("c", values=["y", "n"])]
            ),
            list(zip(
                [42.48, 16.84, 15.23, 23.8],
                ["", "", "", ""],
                "ynyn"
            )))
        table[:, 1] = value
        return table

    @classmethod
    def data_one_column_nans(cls):
        """
        Dataset with two continuous features and one discrete. One continuous
        columns has missing values (NaN).

        Returns
        -------
        data : Orange.data.Table
        """
        return cls.data_one_column_vals(value=np.nan)

    @classmethod
    def data_one_column_infs(cls):
        return cls.data_one_column_vals(value=np.inf)

    @classmethod
    def datasets(cls):
        """
        Yields multiple datasets.

        Returns
        -------
        data : Generator of Orange.data.Table
        """
        ds_cls = Table(cls.path("testing_dataset_cls"))
        ds_reg = Table(cls.path("testing_dataset_reg"))
        for ds in (ds_cls, ds_reg):
            d, a = ds.domain, ds.domain.attributes
            for i in range(0, len(a), 2):
                yield ds.transform(Domain(a[i: i + 2], d.class_vars, d.metas))
            yield ds.transform(Domain(a[:2] + a[8: 10], d.class_vars, d.metas))
        yield cls.missing_data_1()
        yield cls.missing_data_2()
        yield cls.missing_data_3()
        yield cls.data_one_column_nans()
        yield ds_cls
        yield ds_reg


@contextmanager
def open_widget_classes():
    with patch.object(OWWidget, "__init_subclass__"):
        yield



from Orange.classification import CalibratedLearner, ThresholdLearner, \
    NaiveBayesLearner
from Orange.data import Table
from Orange.modelling import Learner
from Orange.widgets import gui
from Orange.widgets.widget import Input
from Orange.widgets.settings import Setting
from Orange.widgets.utils.owlearnerwidget import OWBaseLearner
from Orange.widgets.utils.widgetpreview import WidgetPreview


class OWCalibratedLearner(OWBaseLearner):
    name = "Calibrated Learner"
    description = "Wraps another learner with probability calibration and " \
                  "decision threshold optimization"
    icon = "icons/CalibratedLearner.svg"
    priority = 20
    keywords = ["calibration", "threshold"]

    LEARNER = CalibratedLearner

    SigmoidCalibration, IsotonicCalibration, NoCalibration = range(3)
    CalibrationOptions = ("Sigmoid calibration",
                          "Isotonic calibration",
                          "No calibration")
    CalibrationShort = ("Sigmoid", "Isotonic", "")
    CalibrationMap = {
        SigmoidCalibration: CalibratedLearner.Sigmoid,
        IsotonicCalibration: CalibratedLearner.Isotonic}

    OptimizeCA, OptimizeF1, NoThresholdOptimization = range(3)
    ThresholdOptions = ("Optimize classification accuracy",
                        "Optimize F1 score",
                        "No threshold optimization")
    ThresholdShort = ("CA", "F1", "")
    ThresholdMap = {
        OptimizeCA: ThresholdLearner.OptimizeCA,
        OptimizeF1: ThresholdLearner.OptimizeF1}

    learner_name = Setting("", schema_only=True)
    calibration = Setting(SigmoidCalibration)
    threshold = Setting(OptimizeCA)

    class Inputs(OWBaseLearner.Inputs):
        base_learner = Input("Base Learner", Learner)

    def __init__(self):
        super().__init__()
        self.base_learner = None

    def add_main_layout(self):
        gui.radioButtons(
            self.controlArea, self, "calibration", self.CalibrationOptions,
            box="Probability calibration",
            callback=self.calibration_options_changed)
        gui.radioButtons(
            self.controlArea, self, "threshold", self.ThresholdOptions,
            box="Decision threshold optimization",
            callback=self.calibration_options_changed)

    @Inputs.base_learner
    def set_learner(self, learner):
        self.base_learner = learner
        self._set_default_name()
        self.unconditional_apply()

    def _set_default_name(self):
        if self.base_learner is None:
            self.name = "Calibrated learner"
        else:
            self.name = " + ".join(part for part in (
                self.base_learner.name.title(),
                self.CalibrationShort[self.calibration],
                self.ThresholdShort[self.threshold]) if part)
        self.controls.learner_name.setPlaceholderText(self.name)

    def calibration_options_changed(self):
        self._set_default_name()
        self.apply()

    def create_learner(self):
        class IdentityWrapper(Learner):
            def fit_storage(self, data):
                return self.base_learner.fit_storage(data)

        if self.base_learner is None:
            return None
        learner = self.base_learner
        if self.calibration != self.NoCalibration:
            learner = CalibratedLearner(learner,
                                        self.CalibrationMap[self.calibration])
        if self.threshold != self.NoThresholdOptimization:
            learner = ThresholdLearner(learner,
                                       self.ThresholdMap[self.threshold])
        if self.preprocessors:
            if learner is self.base_learner:
                learner = IdentityWrapper()
            learner.preprocessors = (self.preprocessors, )
        return learner

    def get_learner_parameters(self):
        return (("Calibrate probabilities",
                 self.CalibrationOptions[self.calibration]),
                ("Threshold optimization",
                 self.ThresholdOptions[self.threshold]))


if __name__ == "__main__":  # pragma: no cover
    WidgetPreview(OWCalibratedLearner).run(
        Table("heart_disease"),
        set_learner=NaiveBayesLearner())



# Test methods with long descriptive names can omit docstrings
# pylint: disable=missing-docstring

import unittest

import numpy as np

from Orange.data import Table
from Orange.preprocess import Remove
from Orange.tests import test_filename


class TestRemover(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.test8 = Table(test_filename('datasets/test8.tab'))

    def test_remove(self):
        data = Table("iris")[:5]
        attr_flags = sum([Remove.SortValues,
                          Remove.RemoveConstant,
                          Remove.RemoveUnusedValues])
        class_flags = sum([Remove.SortValues,
                           Remove.RemoveConstant,
                           Remove.RemoveUnusedValues])
        remover = Remove(attr_flags, class_flags)
        new_data = remover(data)
        attr_res, class_res = remover.attr_results, remover.class_results

        self.assertEqual([a.name for a in new_data.domain.attributes],
                         ["sepal length", "sepal width", "petal length"])
        self.assertEqual([c.name for c in new_data.domain.class_vars], [])
        self.assertDictEqual(attr_res,
                             {'removed': 1, 'reduced': 0, 'sorted': 0})
        self.assertDictEqual(class_res,
                             {'removed': 1, 'reduced': 0, 'sorted': 0})

    def test_remove_constant_attr(self):
        data = self.test8
        remover = Remove(Remove.RemoveConstant)
        new_data = remover(data)
        attr_res, class_res = remover.attr_results, remover.class_results

        np.testing.assert_equal(new_data.X, np.hstack((data[:, 1],
                                                       data[:, 3])))
        np.testing.assert_equal(new_data.Y, data.Y)
        self.assertEqual([a.name for a in new_data.domain.attributes],
                         ["c0", "d0"])
        self.assertEqual([c.name for c in new_data.domain.class_vars],
                         ["cl1", "cl0", "cl3", "cl4"])
        self.assertEqual([a.values for a in new_data.domain.attributes
                          if a.is_discrete], [['4', '6']])
        self.assertEqual([c.values for c in new_data.domain.class_vars
                          if c.is_discrete], [['1', '2', '3'], ['2']])
        self.assertDictEqual(attr_res,
                             {'removed': 2, 'reduced': 0, 'sorted': 0})
        self.assertDictEqual(class_res,
                             {'removed': 0, 'reduced': 0, 'sorted': 0})

    def test_remove_constant_class(self):
        data = self.test8
        remover = Remove(class_flags=Remove.RemoveConstant)
        new_data = remover(data)
        attr_res, class_res = remover.attr_results, remover.class_results

        np.testing.assert_equal(new_data.X, data.X)
        np.testing.assert_equal(new_data.Y, np.hstack((data[:, 4],
                                                       data[:, 5])))
        self.assertEqual([a.name for a in new_data.domain.attributes],
                         ["c1", "c0", "d1", "d0"])
        self.assertEqual([c.name for c in new_data.domain.class_vars],
                         ["cl1", "cl0"])
        self.assertEqual([a.values for a in new_data.domain.attributes
                          if a.is_discrete], [['1'], ['4', '6']])
        self.assertEqual([c.values for c in new_data.domain.class_vars
                          if c.is_discrete], [['1', '2', '3']])
        self.assertDictEqual(attr_res,
                             {'removed': 0, 'reduced': 0, 'sorted': 0})
        self.assertDictEqual(class_res,
                             {'removed': 2, 'reduced': 0, 'sorted': 0})

    def test_remove_unused_values_attr(self):
        data = self.test8
        data = data[1:]
        remover = Remove(Remove.RemoveUnusedValues)
        new_data = remover(data)
        attr_res, class_res = remover.attr_results, remover.class_results

        np.testing.assert_equal(new_data.X, data.X)
        np.testing.assert_equal(new_data.Y, data.Y)
        self.assertEqual([a.name for a in new_data.domain.attributes],
                         ["c1", "c0", "d1", "d0"])
        self.assertEqual([c.name for c in new_data.domain.class_vars],
                         ["cl1", "cl0", "cl3", "cl4"])
        self.assertEqual([a.values for a in new_data.domain.attributes
                          if a.is_discrete], [['1'], ['4']])
        self.assertEqual([c.values for c in new_data.domain.class_vars
                          if c.is_discrete], [['1', '2', '3'], ['2']])
        self.assertDictEqual(attr_res,
                             {'removed': 0, 'reduced': 1, 'sorted': 0})
        self.assertDictEqual(class_res,
                             {'removed': 0, 'reduced': 0, 'sorted': 0})

    def test_remove_unused_values_class(self):
        data = self.test8
        data = data[:2]
        remover = Remove(class_flags=Remove.RemoveUnusedValues)
        new_data = remover(data)
        attr_res, class_res = remover.attr_results, remover.class_results

        for i in range(len(data)):
            for j in range(len(data[i])):
                self.assertEqual(new_data[i, j], data[i, j])

        self.assertEqual([a.name for a in new_data.domain.attributes],
                         ["c1", "c0", "d1", "d0"])
        self.assertEqual([c.name for c in new_data.domain.class_vars],
                         ["cl1", "cl0", "cl3", "cl4"])
        self.assertEqual([a.values for a in new_data.domain.attributes
                          if a.is_discrete], [['1'], ['4', '6']])
        self.assertEqual([c.values for c in new_data.domain.class_vars
                          if c.is_discrete], [['2', '3'], ['2']])
        self.assertDictEqual(attr_res,
                             {'removed': 0, 'reduced': 0, 'sorted': 0})
        self.assertDictEqual(class_res,
                             {'removed': 0, 'reduced': 1, 'sorted': 0})

    def test_remove_unused_values_metas(self):
        data = Table(test_filename("datasets/test9.tab"))
        subset = data[:4]
        res = Remove(attr_flags=Remove.RemoveUnusedValues,
                     meta_flags=Remove.RemoveUnusedValues)(subset)

        self.assertEqual(res.domain["b"].values, res.domain["c"].values)
        self.assertEqual(res.domain["d"].values, ["1", "2"])
        self.assertEqual(res.domain["f"].values, ['1', 'hey'])

    def test_remove_unused_values_attr_sparse(self):
        data = self.test8
        data = data[1:].to_sparse()
        remover = Remove(Remove.RemoveUnusedValues)
        new_data = remover(data)
        attr_res = remover.attr_results

        self.assertEqual((new_data.X != data.X).nnz, 0)
        self.assertEqual([a.values for a in new_data.domain.attributes
                          if a.is_discrete], [['1'], ['4']])
        self.assertEqual([c.values for c in new_data.domain.class_vars
                          if c.is_discrete], [['1', '2', '3'], ['2']])
        self.assertDictEqual(attr_res,
                             {'removed': 0, 'reduced': 1, 'sorted': 0})



import unittest
from unittest.mock import patch
from math import sqrt

import numpy as np
from scipy.sparse import csr_matrix

from Orange.data import ContinuousVariable, DiscreteVariable, Domain, Table
from Orange import distance


class CommonTests:
    """Tests applicable to all distance measures"""

    def test_no_data(self):
        """distances return zero-dimensional matrices when no data"""
        n = len(self.data)
        np.testing.assert_almost_equal(
            self.Distance(Table(self.domain)),
            np.zeros((0, 0)))
        np.testing.assert_almost_equal(
            self.Distance(self.data, Table(self.domain)),
            np.zeros((n, 0)))
        np.testing.assert_almost_equal(
            self.Distance(Table(self.domain), self.data),
            np.zeros((0, n)))

    def test_sparse(self):
        """Test sparse support in distances."""
        domain = Domain([ContinuousVariable(c) for c in "abc"])
        dense_data = Table.from_list(
            domain, [[1, 0, 2], [-1, 5, 0], [0, 1, 1], [7, 0, 0]])
        sparse_data = Table(domain, csr_matrix(dense_data.X))

        if not self.Distance.supports_sparse:
            self.assertRaises(TypeError, self.Distance, sparse_data)
        else:
            # check the result is the same for sparse and dense
            dist_dense = self.Distance(dense_data)
            dist_sparse = self.Distance(sparse_data)
            np.testing.assert_allclose(dist_sparse, dist_dense)


class CommonFittedTests(CommonTests):
    """Tests applicable to all distances with fitting"""
    def test_mismatching_attributes(self):
        """distances can't be computed if fit from other attributes"""
        def new_table():
            return Table.from_list(Domain([ContinuousVariable("a")]), [[1]])

        table1 = new_table()
        model = self.Distance().fit(table1)
        self.assertRaises(ValueError, model, new_table())
        self.assertRaises(ValueError, model, table1, new_table())
        self.assertRaises(ValueError, model, new_table(), table1)


class CommonNormalizedTests(CommonFittedTests):
    """Tests applicable to distances the have normalization"""

    def test_zero_variance(self):
        """zero-variance columns have no effect on row distance"""
        assert_almost_equal = np.testing.assert_almost_equal
        normalized = self.Distance(axis=1, normalize=True)
        nonnormalized = self.Distance(axis=1, normalize=False)

        def is_same(d1, d2, fit1=None, fit2=None):
            if fit1 is None:
                fit1 = d1
            if fit2 is None:
                fit2 = d2
            assert_almost_equal(normalized.fit(fit1)(d1),
                                normalized.fit(fit2)(d2))
            assert_almost_equal(nonnormalized.fit(fit1)(d1),
                                nonnormalized.fit(fit2)(d2))

        data = self.cont_data
        n = len(data)
        X = data.X
        domain = Domain(data.domain.attributes + (ContinuousVariable("d"),))
        data_const = Table(domain, np.hstack((X, np.ones((n, 1)))))
        data_nan = Table(domain, np.hstack((X, np.full((n, 1), np.nan))))
        data_nan_1 = Table(domain, np.hstack((X, np.full((n, 1), np.nan))))
        data_nan_1.X[0, -1] = 1
        is_same(data, data_const)
        is_same(data, data_nan)
        is_same(data, data_nan_1)

        # Test whether it's possible to fit with singular data and use the
        # parameters on data where the same column is defined
        is_same(data_nan, data_const, data_const)
        is_same(data_const, data_const, data_nan)

        self.assertRaises(ValueError, self.Distance,
                          data_const, axis=0, normalize=True)
        self.assertRaises(ValueError, self.Distance,
                          data_nan, axis=0, normalize=True)
        self.assertRaises(ValueError, self.Distance,
                          data_nan_1, axis=0, normalize=True)

    def test_mixed_cols(self):
        """distance over columns raises exception for discrete columns"""
        self.assertRaises(ValueError, self.Distance, self.mixed_data, axis=0)
        self.assertRaises(ValueError,
                          self.Distance(axis=0).fit, self.mixed_data)


class FittedDistanceTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.attributes = (
            ContinuousVariable("c1"),
            ContinuousVariable("c2"),
            ContinuousVariable("c3"),
            DiscreteVariable("d1", values=["a", "b"]),
            DiscreteVariable("d2", values=["a", "b", "c", "d"]),
            DiscreteVariable("d3", values=["a", "b", "c"]))
        cls.domain = Domain(cls.attributes)
        cls.cont_domain = Domain(cls.attributes[:3])
        cls.disc_domain = Domain(cls.attributes[3:])

    def setUp(self):
        self.cont_data = Table.from_list(
            self.cont_domain,
            [[1, 3, 2],
             [-1, 5, 0],
             [1, 1, 1],
             [7, 2, 3]])

        self.cont_data2 = Table.from_list(
            self.cont_domain,
            [[2, 1, 3],
             [1, 2, 2]]
        )

        self.disc_data = Table.from_list(
            self.disc_domain,
            [[0, 0, 0],
             [0, 1, 1],
             [1, 3, 1]]
        )

        self.disc_data4 = Table.from_list(
            self.disc_domain,
            [[0, 0, 0],
             [0, 1, 1],
             [0, 1, 1],
             [1, 3, 1]]
        )

        self.mixed_data = self.data = Table.from_numpy(
            self.domain, np.hstack((self.cont_data.X[:3], self.disc_data.X)))



# Correct results in these tests were computed manually or with Excel;
# these are not regression tests
class EuclideanDistanceTest(FittedDistanceTest, CommonNormalizedTests):
    Distance = distance.Euclidean

    def test_euclidean_disc(self):
        assert_almost_equal = np.testing.assert_almost_equal
        data = self.disc_data

        model = distance.Euclidean().fit(data)
        assert_almost_equal(model.dist_missing_disc,
                            [[1/3, 2/3, 1, 1],
                             [2/3, 2/3, 1, 2/3],
                             [2/3, 1/3, 1, 1]])
        assert_almost_equal(model.dist_missing2_disc,
                            [1 - 5/9, 1 - 3/9, 1 - 5/9])

        dist = model(data)
        assert_almost_equal(dist,
                            np.sqrt(np.array([[0, 2, 3],
                                              [2, 0, 2],
                                              [3, 2, 0]])))

        data.X[1, 0] = np.nan
        model = distance.Euclidean().fit(data)
        assert_almost_equal(model.dist_missing_disc,
                            [[1/2, 1/2, 1, 1],
                             [2/3, 2/3, 1, 2/3],
                             [2/3, 1/3, 1, 1]
                            ])
        assert_almost_equal(model.dist_missing2_disc,
                            [1 - 2/4, 1 - 3/9, 1 - 5/9])

        dist = model(data)
        assert_almost_equal(dist,
                            np.sqrt(np.array([[0, 2.5, 3],
                                              [2.5, 0, 1.5],
                                              [3, 1.5, 0]])))

        data.X[0, 0] = np.nan
        model = distance.Euclidean().fit(data)
        assert_almost_equal(model.dist_missing_disc,
                            [[1, 0, 1, 1],
                             [2/3, 2/3, 1, 2/3],
                             [2/3, 1/3, 1, 1]])
        assert_almost_equal(model.dist_missing2_disc,
                            [1 - 1, 1 - 3/9, 1 - 5/9])

        dist = model(data)
        assert_almost_equal(dist,
                            np.sqrt(np.array([[0, 2, 2],
                                              [2, 0, 1],
                                              [2, 1, 0]])))

        data = self.disc_data4
        data.X[:2, 0] = np.nan
        model = distance.Euclidean().fit(data)

        assert_almost_equal(model.dist_missing_disc,
                            [[1/2, 1/2, 1, 1],
                             [3/4, 2/4, 1, 3/4],
                             [3/4, 1/4, 1, 1]])
        assert_almost_equal(model.dist_missing2_disc,
                            [1 - 2/4, 1 - 6/16, 1 - 10/16])

        dist = model(data)
        assert_almost_equal(dist,
                            np.sqrt(np.array([[0, 2.5, 2.5, 2.5],
                                              [2.5, 0, 0.5, 1.5],
                                              [2.5, 0.5, 0, 2],
                                              [2.5, 1.5, 2, 0]])))

    def test_euclidean_cont(self):
        assert_almost_equal = np.testing.assert_almost_equal
        data = self.cont_data

        dist = distance.Euclidean(data, axis=1, normalize=False)
        assert_almost_equal(
            dist,
            np.sqrt(np.array([[0, 12, 5, 38],
                              [12, 0, 21, 82],
                              [5, 21, 0, 41],
                              [38, 82, 41, 0]])))

        data.X[1, 0] = np.nan
        dist = distance.Euclidean(data, axis=1, normalize=False)
        assert_almost_equal(
            dist,
            [[0, 4.472135955, 2.236067977, 6.164414003],
             [4.472135955, 0, 5.385164807, 6.480740698],
             [2.236067977, 5.385164807, 0, 6.403124237],
             [6.164414003, 6.480740698, 6.403124237, 0]])

        data.X[0, 0] = np.nan
        dist = distance.Euclidean(data, axis=1, normalize=False)
        assert_almost_equal(
            dist,
            [[0, 5.099019514, 4.795831523, 4.472135955],
             [5.099019514, 0, 5.916079783, 6],
             [4.795831523, 5.916079783, 0, 6.403124237],
             [4.472135955, 6, 6.403124237, 0]])

    def test_euclidean_cont_normalized(self):
        assert_almost_equal = np.testing.assert_almost_equal
        data = self.cont_data

        model = distance.Euclidean(axis=1, normalize=True).fit(data)
        assert_almost_equal(model.means, [2, 2.75, 1.5])
        assert_almost_equal(model.vars, [9, 2.1875, 1.25])
        assert_almost_equal(model.dist_missing2_cont, [1, 1, 1])

        dist = model(data)
        assert_almost_equal(
            dist,
            [[0, 1.654239383, 1.146423008, 1.621286967],
             [1.654239383, 0, 2.068662631, 3.035242727],
             [1.146423008, 2.068662631, 0, 1.956673562],
             [1.621286967, 3.035242727, 1.956673562, 0]])

        dist = distance.Euclidean(data, axis=1, normalize=True)
        assert_almost_equal(
            dist,
            [[0, 1.654239383, 1.146423008, 1.621286967],
             [1.654239383, 0, 2.068662631, 3.035242727],
             [1.146423008, 2.068662631, 0, 1.956673562],
             [1.621286967, 3.035242727, 1.956673562, 0]])

        data.X[1, 0] = np.nan
        model = distance.Euclidean(axis=1, normalize=True).fit(data)
        assert_almost_equal(model.means, [3, 2.75, 1.5])
        assert_almost_equal(model.vars, [8, 2.1875, 1.25])
        dist = model(data)
        assert_almost_equal(
            dist,
            [[0, 1.806733438, 1.146423008, 1.696635326],
             [1.806733438, 0, 2.192519751, 2.675283697],
             [1.146423008, 2.192519751, 0, 2.019547333],
             [1.696635326, 2.675283697, 2.019547333, 0]])

        data.X[0, 0] = np.nan
        model = distance.Euclidean(axis=1, normalize=True).fit(data)
        assert_almost_equal(model.means, [4, 2.75, 1.5])
        assert_almost_equal(model.vars, [9, 2.1875, 1.25])
        dist = model(data)
        assert_almost_equal(
            dist,
            [[0, 1.874642823, 1.521277659, 1.276154939],
             [1.874642823, 0, 2.248809209, 2.580143961],
             [1.521277659, 2.248809209, 0, 1.956673562],
             [1.276154939, 2.580143961, 1.956673562, 0]])

    def test_euclidean_cols(self):
        assert_almost_equal = np.testing.assert_almost_equal
        data = self.cont_data

        dist = distance.Euclidean(data, axis=0, normalize=False)
        assert_almost_equal(
            dist,
            [[0, 8.062257748, 4.242640687],
             [8.062257748, 0, 5.196152423],
             [4.242640687, 5.196152423, 0]])

        data.X[1, 1] = np.nan
        dist = distance.Euclidean(data, axis=0, normalize=False)
        assert_almost_equal(
            dist,
            [[0, 6.218252702, 4.242640687],
             [6.218252702, 0, 2.581988897],
             [4.242640687, 2.581988897, 0]])

        data.X[1, 0] = np.nan
        dist = distance.Euclidean(data, axis=0, normalize=False)
        assert_almost_equal(
            dist,
            [[0, 6.218252702, 5.830951895],
             [6.218252702, 0, 2.581988897],
             [5.830951895, 2.581988897, 0]])

    def test_euclidean_cols_normalized(self):
        assert_almost_equal = np.testing.assert_almost_equal
        data = self.cont_data

        dist = distance.Euclidean(data, axis=0, normalize=True)
        assert_almost_equal(
            dist,
            [[0, 2.455273959, 0.649839392],
             [2.455273959, 0, 2.473176308],
             [0.649839392, 2.473176308, 0]])

        data.X[1, 1] = np.nan
        dist = distance.Euclidean(data, axis=0, normalize=True)
        assert_almost_equal(
            dist,
            [[0, 2, 0.649839392],
             [2, 0, 1.704275472],
             [0.649839392, 1.704275472, 0]])

        data.X[1, 0] = np.nan
        dist = distance.Euclidean(data, axis=0, normalize=True)
        assert_almost_equal(
            dist,
            [[0, 2, 1.450046001],
             [2, 0, 1.704275472],
             [1.450046001, 1.704275472, 0]])

    def test_euclidean_mixed(self):
        assert_almost_equal = np.testing.assert_almost_equal
        data = self.mixed_data

        model = distance.Euclidean(axis=1, normalize=True).fit(data)

        assert_almost_equal(model.means, [1/3, 3, 1])
        assert_almost_equal(model.vars, [8/9, 8/3, 2/3])
        assert_almost_equal(model.dist_missing_disc,
                            [[1/3, 2/3, 1, 1],
                             [2/3, 2/3, 1, 2/3],
                             [2/3, 1/3, 1, 1]])
        assert_almost_equal(model.dist_missing2_cont, [1, 1, 1])
        assert_almost_equal(model.dist_missing2_disc,
                            [1 - 5/9, 1 - 3/9, 1 - 5/9])
        dist = model(data)
        assert_almost_equal(
            dist,
            [[0, 2.828427125, 2.121320344],
             [2.828427125, 0, 2.828427125],
             [2.121320344, 2.828427125, 0]])

    def test_two_tables(self):
        assert_almost_equal = np.testing.assert_almost_equal

        dist = distance.Euclidean(self.cont_data, self.cont_data2,
                                  normalize=True)
        assert_almost_equal(
            dist,
            [[1.17040218, 0.47809144],
             [2.78516478, 1.96961039],
             [1.28668394, 0.79282497],
             [1.27179413, 1.54919334]])

        model = distance.Euclidean(normalize=True).fit(self.cont_data)
        dist = model(self.cont_data, self.cont_data2)
        assert_almost_equal(
            dist,
            [[1.17040218, 0.47809144],
             [2.78516478, 1.96961039],
             [1.28668394, 0.79282497],
             [1.27179413, 1.54919334]])

        dist = model(self.cont_data2)
        assert_almost_equal(dist, [[0, 0.827119692], [0.827119692, 0]])


class ManhattanDistanceTest(FittedDistanceTest, CommonNormalizedTests):
    Distance = distance.Euclidean

    # The data used for testing Euclidean distances unfortunately yields
    # mads = 1, so we change it a bit
    def setUp(self):
        super().setUp()
        self.cont_data = Table.from_list(
            self.cont_domain,
            [[1, 3, 2],
             [-1, 6, 0],
             [2, 7, 1],
             [7, 1, 3]])

    def test_manhattan_no_data(self):
        np.testing.assert_almost_equal(
            distance.Manhattan(Table(self.domain)),
            np.zeros((0, 0)))
        np.testing.assert_almost_equal(
            distance.Manhattan(self.mixed_data, Table(self.domain)),
            np.zeros((3, 0)))
        np.testing.assert_almost_equal(
            distance.Manhattan(Table(self.domain), self.mixed_data),
            np.zeros((0, 3)))
        self.assertRaises(
            ValueError,
            distance.Manhattan, Table(self.cont_domain), axis=0, normalize=True)

    def test_manhattan_disc(self):
        assert_almost_equal = np.testing.assert_almost_equal
        data = self.disc_data

        model = distance.Manhattan().fit(data)
        assert_almost_equal(model.dist_missing_disc,
                            [[1/3, 2/3, 1, 1],
                             [2/3, 2/3, 1, 2/3],
                             [2/3, 1/3, 1, 1]])
        assert_almost_equal(model.dist_missing2_disc,
                            [1 - 5/9, 1 - 3/9, 1 - 5/9])
        dist = model(data)
        assert_almost_equal(dist,
                            [[0, 2, 3],
                             [2, 0, 2],
                             [3, 2, 0]])

        data.X[1, 0] = np.nan
        model = distance.Manhattan().fit(data)
        assert_almost_equal(model.dist_missing_disc,
                            [[1/2, 1/2, 1, 1],
                             [2/3, 2/3, 1, 2/3],
                             [2/3, 1/3, 1, 1]])
        assert_almost_equal(model.dist_missing2_disc,
                            [1 - 2/4, 1 - 3/9, 1 - 5/9])

        dist = model(data)
        assert_almost_equal(dist,
                            [[0, 2.5, 3],
                             [2.5, 0, 1.5],
                             [3, 1.5, 0]])

        data.X[0, 0] = np.nan
        model = distance.Manhattan().fit(data)
        assert_almost_equal(model.dist_missing_disc,
                            [[1, 0, 1, 1],
                             [2/3, 2/3, 1, 2/3],
                             [2/3, 1/3, 1, 1]])
        assert_almost_equal(model.dist_missing2_disc,
                            [1 - 1, 1 - 3/9, 1 - 5/9])

        dist = model(data)
        assert_almost_equal(dist,
                            [[0, 2, 2],
                             [2, 0, 1],
                             [2, 1, 0]])

        data = self.disc_data4
        data.X[:2, 0] = np.nan
        model = distance.Manhattan().fit(data)
        assert_almost_equal(model.dist_missing_disc,
                            [[1/2, 1/2, 1, 1],
                             [3/4, 2/4, 1, 3/4],
                             [3/4, 1/4, 1, 1]])
        assert_almost_equal(model.dist_missing2_disc,
                            [1 - 2/4, 1 - 6/16, 1 - 10/16])

        dist = model(data)
        assert_almost_equal(dist,
                            [[0, 2.5, 2.5, 2.5],
                             [2.5, 0, 0.5, 1.5],
                             [2.5, 0.5, 0, 2],
                             [2.5, 1.5, 2, 0]])

    def test_manhattan_cont(self):
        assert_almost_equal = np.testing.assert_almost_equal
        data = self.cont_data

        dist = distance.Manhattan(data, axis=1, normalize=False)
        assert_almost_equal(
            dist,
            [[0, 7, 6, 9],
             [7, 0, 5, 16],
             [6, 5, 0, 13],
             [9, 16, 13, 0]])

        data.X[1, 0] = np.nan
        dist = distance.Manhattan(data, axis=1, normalize=False)
        assert_almost_equal(
            dist,
            [[0, 7, 6, 9],
             [7, 0, 3, 14],
             [6, 3, 0, 13],
             [9, 14, 13, 0]])

        data.X[0, 0] = np.nan
        dist = distance.Manhattan(data, axis=1, normalize=False)
        assert_almost_equal(
            dist,
            [[0, 10, 10, 8],
             [10, 0, 7, 13],
             [10, 7, 0, 13],
             [8, 13, 13, 0]])

    def test_manhattan_cont_normalized(self):
        assert_almost_equal = np.testing.assert_almost_equal
        data = self.cont_data

        model = distance.Manhattan(axis=1, normalize=True).fit(data)
        assert_almost_equal(model.medians, [1.5, 4.5, 1.5])
        assert_almost_equal(model.mads, [1.5, 2, 1])
        assert_almost_equal(model.dist_missing2_cont, np.ones(3))

        dist = model(data)
        assert_almost_equal(
            dist,
            [[0, 2.416666667, 1.833333333, 3],
             [2.416666667, 0, 1.75, 5.416666667],
             [1.833333333, 1.75, 0, 4.166666667],
             [3, 5.416666667, 4.166666667, 0]])

        dist = distance.Manhattan(data, axis=1, normalize=True)
        assert_almost_equal(
            dist,
            [[0, 2.416666667, 1.833333333, 3],
             [2.416666667, 0, 1.75, 5.416666667],
             [1.833333333, 1.75, 0, 4.166666667],
             [3, 5.416666667, 4.166666667, 0]])

        data.X[1, 0] = np.nan
        model = distance.Manhattan(axis=1, normalize=True).fit(data)
        assert_almost_equal(model.medians, [2, 4.5, 1.5])
        assert_almost_equal(model.mads, [1, 2, 1])

        dist = model(data)
        assert_almost_equal(
            dist,
            [[0, 2.75, 2, 4],
             [2.75, 0, 1.25, 5.75],
             [2, 1.25, 0, 5],
             [4, 5.75, 5, 0]])

        data.X[0, 0] = np.nan
        model = distance.Manhattan(axis=1, normalize=True).fit(data)
        assert_almost_equal(model.medians, [4.5, 4.5, 1.5])
        assert_almost_equal(model.mads, [2.5, 2, 1])

        dist = model(data)
        assert_almost_equal(
            dist,
            [[0, 2.75, 2.5, 2],
             [2.75, 0, 1.75, 3.75],
             [2.5, 1.75, 0, 3.5],
             [2, 3.75, 3.5, 0]])

    def test_manhattan_cols(self):
        assert_almost_equal = np.testing.assert_almost_equal
        data = self.cont_data

        dist = distance.Manhattan(data, axis=0, normalize=False)
        assert_almost_equal(
            dist,
            [[0, 20, 7],
             [20, 0, 15],
             [7, 15, 0]])

        data.X[1, 1] = np.nan
        dist = distance.Manhattan(data, axis=0, normalize=False)
        assert_almost_equal(
            dist,
            [[0, 19, 7],
             [19, 0, 14],
             [7, 14, 0]])

        data.X[1, 0] = np.nan
        dist = distance.Manhattan(data, axis=0, normalize=False)
        assert_almost_equal(
            dist,
            [[0, 17, 9],
             [17, 0, 14],
             [9, 14, 0]])


    def test_manhattan_cols_normalized(self):
        assert_almost_equal = np.testing.assert_almost_equal
        data = self.cont_data

        dist = distance.Manhattan(data, axis=0, normalize=True)
        assert_almost_equal(
            dist,
            [[0, 4.5833333, 2],
             [4.5833333, 0, 4.25],
             [2, 4.25, 0]])

        data.X[1, 1] = np.nan
        dist = distance.Manhattan(data, axis=0, normalize=True)
        assert_almost_equal(
            dist,
            [[0, 4.6666667, 2],
             [4.6666667, 0, 4],
             [2, 4, 0]])

        data.X[1, 0] = np.nan
        dist = distance.Manhattan(data, axis=0, normalize=True)
        assert_almost_equal(
            dist,
            [[0, 5.5, 4],
             [5.5, 0, 4],
             [4, 4, 0]])

    def test_manhattan_mixed(self):
        assert_almost_equal = np.testing.assert_almost_equal
        data = self.mixed_data

        data.X[2, 0] = 2  # prevent mads[0] = 0
        model = distance.Manhattan(axis=1, normalize=True).fit(data)
        assert_almost_equal(model.medians, [1, 3, 1])
        assert_almost_equal(model.mads, [1, 2, 1])
        assert_almost_equal(model.dist_missing_disc,
                            [[1/3, 2/3, 1, 1],
                             [2/3, 2/3, 1, 2/3],
                             [2/3, 1/3, 1, 1]])
        assert_almost_equal(model.dist_missing2_disc,
                            [1 - 5/9, 1 - 3/9, 1 - 5/9])

        dist = model(data)
        assert_almost_equal(
            dist,
            [[0, 4.5, 4.5],
             [4.5, 0, 5],
             [4.5, 5, 0]])

    def test_two_tables(self):
        assert_almost_equal = np.testing.assert_almost_equal

        dist = distance.Manhattan(self.cont_data, self.cont_data2,
                                  normalize=True)
        assert_almost_equal(
            dist,
            [[1.3333333, 0.25],
             [3.75, 2.6666667],
             [2.5, 2.0833333],
             [1.6666667, 2.75]])

        model = distance.Manhattan(normalize=True).fit(self.cont_data)
        dist = model(self.cont_data, self.cont_data2)
        assert_almost_equal(
            dist,
            [[1.3333333, 0.25],
             [3.75, 2.6666667],
             [2.5, 2.0833333],
             [1.6666667, 2.75]])

        dist = model(self.cont_data2)
        assert_almost_equal(dist, [[0, 1.083333333], [1.083333333, 0]])

    def test_manhattan_mixed_cols(self):
        self.assertRaises(ValueError,
                          distance.Manhattan, self.mixed_data, axis=0)
        self.assertRaises(ValueError,
                          distance.Manhattan(axis=0).fit, self.mixed_data)


class CosineDistanceTest(FittedDistanceTest, CommonFittedTests):
    Distance = distance.Cosine

    def test_no_data(self):
        with patch("warnings.warn") as warn:
            super().test_no_data()
            self.assertEqual(warn.call_args[0],
                             ("Mean of empty slice", RuntimeWarning))

    def test_cosine_disc(self):
        assert_almost_equal = np.testing.assert_almost_equal
        data = self.disc_data
        data.X = np.array([[1, 0, 0],
                           [0, 1, 1],
                           [1, 3, 0]], dtype=float)

        model = distance.Cosine().fit(data)
        assert_almost_equal(model.means, [2 / 3, 2 / 3, 1 / 3])

        dist = model(data)
        assert_almost_equal(dist, 1 - np.array([[1, 0, 1 / sqrt(2)],
                                                [0, 1, 0.5],
                                                [1 / sqrt(2), 0.5, 1]]))

        data.X[1, 1] = np.nan
        model = distance.Cosine().fit(data)
        assert_almost_equal(model.means, [2 / 3, 1 / 2, 1 / 3])
        dist = model(data)
        assert_almost_equal(
            dist,
            1 - np.array([[1, 0, 1 / sqrt(2)],
                          [0, 1, 0.5 / sqrt(1.25) / sqrt(2)],
                          [1 / sqrt(2), 0.5 / sqrt(1.25) / sqrt(2), 1]]))

        data.X = np.array([[1, 0, 0],
                           [0, np.nan, 1],
                           [1, np.nan, 1],
                           [1, 3, 1]])
        model = distance.Cosine().fit(data)
        dist = model(data)
        assert_almost_equal(model.means, [0.75, 0.5, 0.75])
        assert_almost_equal(dist, [[0, 1, 0.333333333, 0.422649731],
                                   [1, 0, 0.254644008, 0.225403331],
                                   [0.333333333, 0.254644008, 0, 0.037749551],
                                   [0.422649731, 0.225403331, 0.037749551, 0]])

    def test_cosine_cont(self):
        assert_almost_equal = np.testing.assert_almost_equal
        data = self.cont_data

        dist = distance.Cosine(data, axis=1)
        assert_almost_equal(
            dist,
            [[0, 0.266200614, 0.0741799, 0.355097978],
             [0.266200614, 0, 0.547089186, 0.925279678],
             [0.0741799, 0.547089186, 0, 0.12011731],
             [0.355097978, 0.925279678, 0.12011731, 0]]
        )

        data.X[1, 0] = np.nan
        dist = distance.Cosine(data, axis=1)
        assert_almost_equal(
            dist,
            [[0, 0.174971353, 0.0741799, 0.355097978],
             [0.174971353, 0, 0.207881966, 0.324809395],
             [0.0741799, 0.207881966, 0, 0.12011731],
             [0.355097978, 0.324809395, 0.12011731, 0]])

        data.X[0, 0] = np.nan
        dist = distance.Cosine(data, axis=1)
        assert_almost_equal(
            dist,
            [[0, 0.100977075, 0.035098719, 0.056666739],
             [0.100977075, 0, 0.188497329, 0.246304671],
             [0.035098719, 0.188497329, 0, 0.12011731],
             [0.056666739, 0.246304671, 0.12011731, 0]])

    def test_cosine_mixed(self):
        assert_almost_equal = np.testing.assert_almost_equal
        data = self.mixed_data
        data.X = np.array([[1, 3, 2, 1, 0, 0],
                           [-1, 5, 0, 0, 1, 1],
                           [1, 1, 1, 1, 3, 0]], dtype=float)

        model = distance.Cosine(axis=1).fit(data)
        assert_almost_equal(model.means, [1/3, 3, 1, 2/3, 2/3, 1/3])
        dist = model(data)
        assert_almost_equal(
            dist,
            [[0, 0.316869949, 0.191709623],
             [0.316869949, 0, 0.577422873],
             [0.191709623, 0.577422873, 0]])

    def test_two_tables(self):
        assert_almost_equal = np.testing.assert_almost_equal
        self.cont_data.X[1, 0] = np.nan
        self.cont_data2.X[1, 0] = np.nan

        dist = distance.Cosine(self.cont_data, self.cont_data2)
        assert_almost_equal(
            dist,
            [[0.2142857, 0.1573352],
             [0.4958158, 0.2097042],
             [0.0741799, 0.0198039],
             [0.1514447, 0.0451363]])

        model = distance.Cosine().fit(self.cont_data)
        dist = model(self.cont_data, self.cont_data2)
        assert_almost_equal(
            dist,
            [[0.2142857, 0.1573352],
             [0.4958158, 0.2097042],
             [0.0741799, 0.0198039],
             [0.1514447, 0.0451363]])

        dist = model(self.cont_data2)
        assert_almost_equal(dist, [[0, 0.092514787], [0.092514787, 0]])

    def test_cosine_cols(self):
        assert_almost_equal = np.testing.assert_almost_equal
        data = self.cont_data

        dist = distance.Cosine(data, axis=0, normalize=False)
        assert_almost_equal(
            dist,
            [[0, 0.711324865, 0.11050082],
             [0.711324865, 0, 0.44365136],
             [0.11050082, 0.44365136, 0]])

        data.X[1, 1] = np.nan
        dist = distance.Cosine(data, axis=0, normalize=False)
        assert_almost_equal(
            dist,
            [[0, 0.47702364, 0.11050082],
             [0.47702364, 0, 0.181076975],
             [0.11050082, 0.181076975, 0]])

        data.X[1, 0] = np.nan
        data.X[1, 2] = 2
        dist = distance.Cosine(data, axis=0, normalize=False)
        assert_almost_equal(
            dist,
            [[0, 0.269703257, 0.087129071],
             [0.269703257, 0, 0.055555556],
             [0.087129071, 0.055555556, 0]])


class JaccardDistanceTest(unittest.TestCase, CommonFittedTests):
    Distance = distance.Jaccard

    def setUp(self):
        self.domain = Domain([DiscreteVariable(c) for c in "abc"])
        self.data = Table(
            self.domain,
            [[0, 1, 1],
             [1, 1, 1],
             [1, 0, 1],
             [1, 0, 0]])

    def test_jaccard_rows(self):
        assert_almost_equal = np.testing.assert_almost_equal

        model = distance.Jaccard().fit(self.data)
        assert_almost_equal(model.ps, [0.75, 0.5, 0.75])
        assert_almost_equal(
            model(self.data),
            1 - np.array([[1, 2/3, 1/3, 0],
                          [2/3, 1, 2/3, 1/3],
                          [1/3, 2/3, 1, 1/2],
                          [0, 1/3, 1/2, 1]]))

        X = self.data.X
        X[1, 0] = X[2, 0] = X[3, 1] = np.nan
        model = distance.Jaccard().fit(self.data)
        assert_almost_equal(model.ps, np.array([0.5, 2/3, 0.75]))

        # pylint: disable=bad-whitespace
        assert_almost_equal(
            model(self.data),
            1 - np.array([[      1,        2 / 2.5,        1 / 2.5,        2/3 / 3],
                          [2 / 2.5,              1,    1.25 / 2.75,  (1/2+2/3) / 3],
                          [1 / 2.5,    1.25 / 2.75,              1,  0.5 / (2+2/3)],
                          [2/3 / 3,  (1/2+2/3) / 3,  0.5 / (2+2/3),            1]]))

    def test_jaccard_cols(self):
        assert_almost_equal = np.testing.assert_almost_equal
        model = distance.Jaccard(axis=0).fit(self.data)
        assert_almost_equal(model.ps, [0.75, 0.5, 0.75])
        assert_almost_equal(
            model(self.data),
            1 - np.array([[1, 1/4, 1/2],
                          [1/4, 1, 2/3],
                          [1/2, 2/3, 1]]))

        self.data.X = np.array([[0, 1, 1],
                                [np.nan, np.nan, 1],
                                [np.nan, 0, 1],
                                [1, 1, 0]])
        model = distance.Jaccard(axis=0).fit(self.data)
        assert_almost_equal(model.ps, [0.5, 2/3, 0.75])
        assert_almost_equal(
            model(self.data),
            1 - np.array([[1, 0.4, 0.25],
                          [0.4, 1, 5/12],
                          [0.25, 5/12, 1]]))

    def test_zero_instances(self):
        "Test all zero instances"
        domain = Domain([ContinuousVariable(c) for c in "abc"])
        dense_data = Table.from_list(
            domain, [[0, 0, 0], [0, 0, 0], [0, 0, 1]])
        sparse_data = Table(domain, csr_matrix(dense_data.X))
        dist_dense = self.Distance(dense_data)
        dist_sparse = self.Distance(sparse_data)

        self.assertEqual(dist_dense[0][1], 0)
        self.assertEqual(dist_sparse[0][1], 0)
        self.assertEqual(dist_dense[0][2], 1)
        self.assertEqual(dist_sparse[0][2], 1)


class HammingDistanceTest(FittedDistanceTest):
    Distance = distance.Hamming

    def test_hamming_col(self):
        np.testing.assert_almost_equal(
            self.Distance(self.disc_data, axis=0),
            [[0, 2 / 3, 1 / 3],
             [2 / 3, 0, 1 / 3],
             [1 / 3, 1 / 3, 0]])

    def test_hamming_row(self):
        np.testing.assert_almost_equal(
            self.Distance(self.disc_data),
            [[0, 2 / 3, 1],
             [2 / 3, 0, 2 / 3],
             [1, 2 / 3, 0]]
        )

    def test_hamming_row_secondary_data(self):
        np.testing.assert_almost_equal(
            self.Distance(self.disc_data, self.disc_data[:2]),
            [[0, 2 / 3],
             [2 / 3, 0],
             [1, 2 / 3]])


if __name__ == "__main__":
    unittest.main()



import unittest
from Orange.widgets.settings import Context, migrate_str_to_variable


class MigrationsTestCase(unittest.TestCase):
    def test_migrate_str_to_variable(self):
        values = dict(foo=("foo", 1), baz=("baz", 2), qux=("qux", 102), bar=13)

        context = Context(values=values.copy())
        migrate_str_to_variable(context)
        self.assertDictEqual(
            context.values,
            dict(foo=("foo", 101), baz=("baz", 102), qux=("qux", 102), bar=13))

        context = Context(values=values.copy())
        migrate_str_to_variable(context, ("foo", "qux"))
        self.assertDictEqual(
            context.values,
            dict(foo=("foo", 101), baz=("baz", 2), qux=("qux", 102), bar=13))

        context = Context(values=values.copy())
        migrate_str_to_variable(context, "foo")
        self.assertDictEqual(
            context.values,
            dict(foo=("foo", 101), baz=("baz", 2), qux=("qux", 102), bar=13))

        self.assertRaises(KeyError, migrate_str_to_variable, context, "quuux")



from functools import wraps

from Orange.widgets.utils.messages import UnboundMsg
from Orange.data import Table
from Orange.data.sql.table import SqlTable, AUTO_DL_LIMIT

_download_sql_data = UnboundMsg(
    "Download (and sample if necessary) the SQL data first")


def check_sql_input(f):
    """
    Wrapper for widget's set_data method that first checks if the input
    is a SqlTable and:
    - if small enough, download all data and convert to Table
    - for large sql tables, show an error

    :param f: widget's `set_data` method to wrap
    :return: wrapped method that handles SQL data inputs
    """
    @wraps(f)
    def new_f(widget, data, *args, **kwargs):
        widget.Error.add_message("download_sql_data", _download_sql_data)
        widget.Error.download_sql_data.clear()
        if isinstance(data, SqlTable):
            if data.approx_len() < AUTO_DL_LIMIT:
                data = Table(data)
            else:
                widget.Error.download_sql_data()
                data = None
        return f(widget, data, *args, **kwargs)

    return new_f



import unittest

from Orange.data import Table
from Orange.ensembles.stack import StackedFitter
from Orange.evaluation import CA, CrossValidation, MSE
from Orange.modelling import KNNLearner, TreeLearner


class TestStackedFitter(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.iris = Table('iris')
        cls.housing = Table('housing')

    def test_classification(self):
        sf = StackedFitter([TreeLearner(), KNNLearner()])
        results = CrossValidation(self.iris, [sf], k=3)
        ca = CA(results)
        self.assertGreater(ca, 0.9)

    def test_regression(self):
        sf = StackedFitter([TreeLearner(), KNNLearner()])
        results = CrossValidation(self.housing[:50],
                                  [sf, TreeLearner(), KNNLearner()], k=3,
                                  random_state=0)
        mse = MSE()(results)
        self.assertLess(mse[0], mse[1])
        self.assertLess(mse[0], mse[2])



import unittest
from unittest.mock import patch

import numpy as np

from Orange.evaluation.testing import Results
from Orange.evaluation.performance_curves import Curves


# Test data and sensitivity/specificity are taken from
# Tom Fawcett: An introduction to ROC analysis, with one true positive instance
# removed, so that the number of positive and negative does not match

class TestCurves(unittest.TestCase):
    def setUp(self):
        n, p = (0, 1)
        self.data = np.array([
            (p, .8), (n, .7), (p, .6), (p, .55), (p, .54), (n, .53),
            (n, .52), (p, .51), (n, .505), (p, .4), (n, .39), (p, .38),
            (n, .37), (n, .36), (n, .35), (p, .34), (n, .33), (p, .30), (n, .1)
        ])

    def test_curves(self):
        np.random.shuffle(self.data)
        ytrue, probs = self.data.T
        curves = Curves(ytrue, probs)

        tn = np.array(
            [0, 1, 1, 2, 2, 3, 4, 5, 5, 6, 6, 7, 7, 8, 9, 9, 9, 9, 10, 10])
        np.testing.assert_equal(curves.tn, tn)
        np.testing.assert_equal(curves.fp, 10 - tn)
        np.testing.assert_almost_equal(curves.specificity(), tn / 10)

        tp = np.array(
            [9, 9, 8, 8, 7, 7, 7, 7, 6, 6, 5, 5, 4, 4, 4, 3, 2, 1, 1, 0])
        np.testing.assert_equal(curves.tp, tp)
        np.testing.assert_equal(curves.fn, 9 - tp)
        np.testing.assert_almost_equal(curves.sensitivity(), tp / 9)

        np.testing.assert_almost_equal(
            curves.ca(),
            np.array([9, 10, 9, 10, 9, 10, 11, 12, 11, 12, 11, 12, 11, 12,
                      13, 12, 11, 10, 11, 10]) / 19)

        precision = np.array(
            [9 / 19, 9 / 18, 8 / 17, 8 / 16, 7 / 15, 7 / 14, 7 / 13,
             7 / 12, 6 / 11, 6 / 10, 5 / 9, 5 / 8, 4 / 7, 4 / 6,
             4 / 5, 3 / 4, 2 / 3, 1 / 2, 1 / 1, 1])
        np.testing.assert_almost_equal(curves.precision(), precision)
        np.testing.assert_almost_equal(curves.recall(), tp / 9)

        np.testing.assert_almost_equal(curves.ppv(), precision)
        np.testing.assert_almost_equal(
            curves.npv(),
            np.array([1, 1 / 1, 1 / 2, 2 / 3, 2 / 4, 3 / 5, 4 / 6, 5 / 7,
                      5 / 8, 6 / 9, 6 / 10, 7 / 11, 7 / 12, 8 / 13, 9 / 14,
                      9 / 15, 9 / 16, 9 / 17, 10 / 18, 10 / 19]))

        np.testing.assert_almost_equal(curves.tpr(), tp / 9)
        np.testing.assert_almost_equal(curves.fpr(), (10 - tn) / 10)

    @patch("Orange.evaluation.performance_curves.Curves.__init__",
           return_value=None)
    def test_curves_from_results(self, init):
        res = Results()
        ytrue, probs = self.data.T
        res.actual = ytrue.astype(float)
        res.probabilities = np.vstack((1 - probs, probs)).T.reshape(1, -1, 2)
        Curves.from_results(res)
        cytrue, cprobs = init.call_args[0]
        np.testing.assert_equal(cytrue, ytrue)
        np.testing.assert_equal(cprobs, probs)

        Curves.from_results(res, target_class=0)
        cytrue, cprobs = init.call_args[0]
        np.testing.assert_equal(cytrue, 1 - ytrue)
        np.testing.assert_equal(cprobs, 1 - probs)

        res.actual = ytrue.astype(float)
        res.probabilities = np.random.random((2, 19, 2))
        res.probabilities[1] = np.vstack((1 - probs, probs)).T

        Curves.from_results(res, model_index=1)
        cytrue, cprobs = init.call_args[0]
        np.testing.assert_equal(cytrue, ytrue)
        np.testing.assert_equal(cprobs, probs)

        self.assertRaises(ValueError, Curves.from_results, res)

        ytrue[ytrue == 0] = 2 * (np.arange(10) % 2)
        res.actual = ytrue.astype(float)
        res.probabilities = np.random.random((2, 19, 3))
        res.probabilities[1] = np.vstack(
            ((1 - probs) / 3, probs, (1 - probs) * 2 / 3)).T

        Curves.from_results(res, model_index=1, target_class=1)
        cytrue, cprobs = init.call_args[0]
        np.testing.assert_equal(cytrue, ytrue == 1)
        np.testing.assert_equal(cprobs, probs)

        Curves.from_results(res, model_index=1, target_class=0)
        cytrue, cprobs = init.call_args[0]
        np.testing.assert_equal(cytrue, ytrue == 0)
        np.testing.assert_equal(cprobs, (1 - probs) / 3)

        Curves.from_results(res, model_index=1, target_class=2)
        cytrue, cprobs = init.call_args[0]
        np.testing.assert_equal(cytrue, ytrue == 2)
        np.testing.assert_equal(cprobs, (1 - probs) * 2 / 3)

        self.assertRaises(ValueError, Curves.from_results, res, model_index=1)

    @patch("Orange.evaluation.performance_curves.Curves.__init__",
           return_value=None)
    def test_curves_from_results_nans(self, init):
        res = Results()
        ytrue, probs = self.data.T
        ytrue[0] = np.nan
        probs[-1] = np.nan
        res.actual = ytrue.astype(float)
        res.probabilities = np.vstack((1 - probs, probs)).T.reshape(1, -1, 2)
        Curves.from_results(res)
        cytrue, cprobs = init.call_args[0]
        np.testing.assert_equal(cytrue, ytrue[1:-1])
        np.testing.assert_equal(cprobs, probs[1:-1])



"""Tree adapter class for sklearn trees."""
from collections import OrderedDict
import random

import numpy as np
from Orange.widgets.visualize.utils.tree.treeadapter import BaseTreeAdapter

from Orange.misc.cache import memoize_method
from Orange.preprocess.transformation import Indicator
from Orange.widgets.visualize.utils.tree.rules import (
    DiscreteRule,
    ContinuousRule
)


class SklTreeAdapter(BaseTreeAdapter):
    """Sklear Tree Adapter.

    An abstraction on top of the scikit learn classification tree.

    Parameters
    ----------
    model : SklModel
        An sklearn tree model instance.

    """

    NO_CHILD = -1
    FEATURE_UNDEFINED = -2

    def __init__(self, model):
        super().__init__(model)
        self._tree = model.skl_model.tree_
        self._all_leaves = None

    @memoize_method(maxsize=1024)
    def weight(self, node):
        return self.num_samples(node) / self.num_samples(self.parent(node))

    def num_samples(self, node):
        return self._tree.n_node_samples[node]

    @memoize_method(maxsize=1024)
    def parent(self, node):
        for children in (self._tree.children_left, self._tree.children_right):
            try:
                return (children == node).nonzero()[0][0]
            except IndexError:
                continue
        return self.ROOT_PARENT

    def has_children(self, node):
        return (self._tree.children_left[node] != self.NO_CHILD or
                self._tree.children_right[node] != self.NO_CHILD)

    def children(self, node):
        if self.has_children(node):
            return self.__left_child(node), self.__right_child(node)
        return ()

    def __left_child(self, node):
        return self._tree.children_left[node]

    def __right_child(self, node):
        return self._tree.children_right[node]

    def reverse_children(self, node):
        self._tree.children_left[node], self._tree.children_right[node] = \
            self._tree.children_right[node], self._tree.children_left[node]

    def shuffle_children(self):
        for i in range(self.num_nodes):
            if random.randrange(2) == 0:
                self.reverse_children(i)

    @memoize_method(maxsize=1024)
    def get_distribution(self, node):
        value = self._tree.value[node]
        # If regression tree, we have to compute variance by hand, we can
        # detect this because you can't have classification trees when there's
        # only one class
        if value.shape[1] == 1:
            var = np.var(self.get_instances_in_nodes(node).Y)
            variances = np.array([(var * np.ones(value.shape[0]))]).T
            value = np.hstack((value, variances))
        return value

    def get_impurity(self, node):
        return self._tree.impurity[node]

    @property
    def max_depth(self):
        return self._tree.max_depth

    @property
    def num_nodes(self):
        return self._tree.node_count

    @property
    def root(self):
        return 0

    @memoize_method(maxsize=1024)
    def rules(self, node):
        if node != self.root:
            parent = self.parent(node)
            # Convert the parent list of rules into an ordered dict
            pr = OrderedDict([(r.attr_name, r) for r in self.rules(parent)])

            parent_attr = self.attribute(parent)
            # Get the parent attribute type
            parent_attr_cv = parent_attr.compute_value

            is_left_child = self.__left_child(parent) == node

            # The parent split variable is discrete
            if isinstance(parent_attr_cv, Indicator) and \
                    hasattr(parent_attr_cv.variable, 'values'):
                values = parent_attr_cv.variable.values
                attr_name = parent_attr_cv.variable.name
                eq = not is_left_child * (len(values) != 2)
                value = values[abs(parent_attr_cv.value -
                                   is_left_child * (len(values) == 2))]
                new_rule = DiscreteRule(attr_name, eq, value)
                # Since discrete variables should appear in their own lines
                # they must not be merged, so the dict key is set with the
                # value, so the same keys can exist with different values
                # e.g. #legs ≠ 2 and #legs ≠ 4
                attr_name = attr_name + '_' + value
            # The parent split variable is continuous
            else:
                attr_name = parent_attr.name
                sign = not is_left_child
                value = self._tree.threshold[self.parent(node)]
                new_rule = ContinuousRule(attr_name, sign, value,
                                          inclusive=is_left_child)

            # Check if a rule with that attribute exists
            if attr_name in pr:
                pr[attr_name] = pr[attr_name].merge_with(new_rule)
                pr.move_to_end(attr_name)
            else:
                pr[attr_name] = new_rule

            return list(pr.values())
        else:
            return []

    def short_rule(self, node):
        return self.rules(node)[0].description

    def attribute(self, node):
        feature_idx = self.splitting_attribute(node)
        if feature_idx == self.FEATURE_UNDEFINED:
            return None
        return self.domain.attributes[self.splitting_attribute(node)]

    def splitting_attribute(self, node):
        return self._tree.feature[node]

    @memoize_method(maxsize=1024)
    def leaves(self, node):
        start, stop = self._subnode_range(node)
        if start == stop:
            # leaf
            return np.array([node], dtype=int)
        else:
            is_leaf = self._tree.children_left[start:stop] == self.NO_CHILD
            assert np.flatnonzero(is_leaf).size > 0
            return start + np.flatnonzero(is_leaf)

    def _subnode_range(self, node):
        """
        Get the range of indices where there are subnodes of the given node.

        See Also
        --------
        Orange.widgets.model.owclassificationtreegraph.OWTreeGraph
        """

        def find_largest_idx(n):
            """It is necessary to locate the node with the largest index in the
            children in order to get a good range. This is necessary with trees
            that are not right aligned, which can happen when visualising
            random forest trees."""
            if self._tree.children_left[n] == self.NO_CHILD:
                return n

            l_node = find_largest_idx(self._tree.children_left[n])
            r_node = find_largest_idx(self._tree.children_right[n])

            return max(l_node, r_node)

        right = left = node
        if self._tree.children_left[left] == self.NO_CHILD:
            assert self._tree.children_right[node] == self.NO_CHILD
            return node, node
        else:
            left = self._tree.children_left[left]
            right = find_largest_idx(right)

            return left, right + 1

    def get_samples_in_leaves(self):
        """Get an array of instance indices that belong to each leaf.

        For a given dataset X, separate the instances out into an array, so
        they are grouped together based on what leaf they belong to.

        Examples
        --------
        Given a tree with two leaf nodes ( A <- R -> B ) and the dataset X =
        [ 10, 20, 30, 40, 50, 60 ], where 10, 20 and 40 belong to leaf A, and
        the rest to leaf B, the following structure will be returned (where
        array is the numpy array):
        [array([ 0, 1, 3 ]), array([ 2, 4, 5 ])]

        The first array represents the indices of the values that belong to the
        first leaft, so calling X[ 0, 1, 3 ] = [ 10, 20, 40 ]

        Parameters
        ----------
        data
            A matrix containing the data instances.

        Returns
        -------
        np.array
            The indices of instances belonging to a given leaf.

        """

        def assign(node_id, indices):
            if self._tree.children_left[node_id] == self.NO_CHILD:
                return [indices]
            else:
                feature_idx = self._tree.feature[node_id]
                thresh = self._tree.threshold[node_id]

                column = self.instances_transformed.X[indices, feature_idx]
                leftmask = column <= thresh
                leftind = assign(self._tree.children_left[node_id],
                                 indices[leftmask])
                rightind = assign(self._tree.children_right[node_id],
                                  indices[~leftmask])
                return list.__iadd__(leftind, rightind)

        if self._all_leaves is not None:
            return self._all_leaves

        n, _ = self.instances.X.shape

        items = np.arange(n, dtype=int)
        leaf_indices = assign(0, items)
        self._all_leaves = leaf_indices
        return leaf_indices

    def get_instances_in_nodes(self, nodes):
        if not isinstance(nodes, (list, tuple)):
            nodes = [nodes]

        node_leaves = [self.leaves(n) for n in nodes]
        if len(node_leaves) > 0:
            # get the leaves of the selected tree node
            node_leaves = np.unique(np.hstack(node_leaves))

            all_leaves = self.leaves(self.root)

            indices = np.searchsorted(all_leaves, node_leaves)
            # all the leaf samples for each leaf
            leaf_samples = self.get_samples_in_leaves()
            # filter out the leaf samples array that are not selected
            leaf_samples = [leaf_samples[i] for i in indices]
            indices = np.hstack(leaf_samples)
        else:
            indices = []

        return self.instances[indices] if len(indices) else None

    def get_indices(self, nodes):
        if not isinstance(nodes, (list, tuple)):
            nodes = [nodes]

        node_leaves = [self.leaves(n) for n in nodes]
        if len(node_leaves) > 0:
            # get the leaves of the selected tree node
            node_leaves = np.unique(np.hstack(node_leaves))

            all_leaves = self.leaves(self.root)

            return np.searchsorted(all_leaves, node_leaves)
        return []



# Test methods with long descriptive names can omit docstrings
# pylint: disable=missing-docstring

import unittest
from unittest.mock import MagicMock, patch
import itertools

import numpy as np

from Orange.data import Table, Domain, ContinuousVariable, DiscreteVariable
from Orange.data.filter import \
    FilterContinuous, FilterDiscrete, FilterString, Values, HasClass, \
    IsDefined, SameValue, Random, ValueFilter, FilterStringList, FilterRegex
from Orange.tests import test_filename

NIMOCK = MagicMock(side_effect=NotImplementedError())


class TestFilterValues(unittest.TestCase):
    def setUp(self):
        self.iris = Table('iris')

    @patch("Orange.data.Table._filter_values", NIMOCK)
    def test_values(self):
        vs = self.iris.domain.variables
        f1 = FilterContinuous(vs[0], FilterContinuous.Less, 5)
        f2 = FilterContinuous(vs[1], FilterContinuous.Greater, 3)
        f3 = FilterDiscrete(vs[4], [2])
        f12 = Values([f1, f2], conjunction=False, negate=True)
        f123 = Values([f12, f3])
        d12 = f12(self.iris)
        d123 = f123(self.iris)
        self.assertGreater(len(d12), len(d123))
        self.assertTrue((d123.X[:, 0] >= 5).all())
        self.assertTrue((d123.X[:, 1] <= 3).all())
        self.assertTrue((d123.Y == 2).all())
        self.assertEqual(len(d123),
                         (~((self.iris.X[:, 0] < 5) | (self.iris.X[:, 1] > 3)) &
                          (self.iris.Y == 2)).sum())


class TestIsDefinedFilter(unittest.TestCase):
    def setUp(self):
        self.table = Table(test_filename('datasets/imports-85.tab'))
        self.n_missing = 46
        self.assertTrue(self.table.has_missing())

    def test_is_defined_filter_table(self):
        filter_ = IsDefined()
        without_missing = filter_(self.table)
        self.assertEqual(len(without_missing),
                         len(self.table) - self.n_missing)
        self.assertFalse(without_missing.has_missing())

        filter_ = IsDefined(negate=True)
        just_missing = filter_(self.table)
        self.assertEqual(len(just_missing), self.n_missing)
        self.assertTrue(just_missing.has_missing())

    def test_is_defined_filter_instance(self):
        instance_with_missing = self.table[0]
        instance_without_missing = self.table[3]

        filter_ = IsDefined()
        self.assertFalse(filter_(instance_with_missing))
        self.assertTrue(filter_(instance_without_missing))

        filter_ = IsDefined(negate=True)
        self.assertTrue(filter_(instance_with_missing))
        self.assertFalse(filter_(instance_without_missing))

    @patch('Orange.data.Table._filter_is_defined', NIMOCK)
    def test_is_defined_filter_not_implemented(self):
        self.test_is_defined_filter_table()


class TestHasClassFilter(unittest.TestCase):
    def setUp(self):
        self.table = Table(test_filename('datasets/imports-85.tab'))
        self.n_missing = 4
        self.assertTrue(self.table.has_missing_class())

    def test_has_class_filter_table(self):
        filter_ = HasClass()
        with_class = filter_(self.table)
        self.assertEqual(len(with_class),
                         len(self.table) - self.n_missing)
        self.assertFalse(with_class.has_missing_class())

        filter_ = HasClass(negate=True)
        without_class = filter_(self.table)
        self.assertEqual(len(without_class), self.n_missing)
        self.assertTrue(without_class.has_missing_class())

    def test_has_class_multiclass(self):
        domain = Domain([DiscreteVariable("x", values="01")],
                        [DiscreteVariable("y1", values="01"),
                         DiscreteVariable("y2", values="01")])
        table = Table(domain, [[0, 1, np.nan],
                               [1, np.nan, 0],
                               [1, 0, 1],
                               [1, np.nan, np.nan]])
        table = HasClass()(table)
        self.assertTrue(not np.isnan(table).any())
        self.assertEqual(table.domain, domain)
        self.assertEqual(len(table), 1)

    def test_has_class_filter_instance(self):
        class_missing = self.table[9]
        class_present = self.table[0]

        filter_ = HasClass()
        self.assertFalse(filter_(class_missing))
        self.assertTrue(filter_(class_present))

        filter_ = HasClass(negate=True)
        self.assertTrue(filter_(class_missing))
        self.assertFalse(filter_(class_present))

    @patch('Orange.data.Table._filter_has_class', NIMOCK)
    def test_has_class_filter_not_implemented(self):
        self.test_has_class_filter_table()


class TestFilterContinuous(unittest.TestCase):
    def setUp(self):
        self.domain = Domain([ContinuousVariable(x) for x in "abcd"])
        self.inst = Table(self.domain, np.array([[0.1, 0.2, 0.3, np.nan]]))[0]

    def test_min(self):
        flt = FilterContinuous(1, FilterContinuous.Between, 1, 2)
        self.assertEqual(flt.min, 1)
        self.assertEqual(flt.max, 2)
        self.assertEqual(flt.ref, 1)

        flt.ref = 0
        self.assertEqual(flt.min, 0)

        flt.min = -1
        self.assertEqual(flt.ref, -1)

        self.assertRaises(
            TypeError,
            FilterContinuous, 1, FilterContinuous.Equal, 0, c=12)
        self.assertRaises(
            TypeError,
            FilterContinuous, 1, FilterContinuous.Equal, 0, min=5, c=12)

        flt = FilterContinuous(1, FilterContinuous.Between, min=1, max=2)
        self.assertEqual(flt.ref, 1)

    def test_operator(self):
        inst = self.inst
        flt = FilterContinuous
        self.assertTrue(flt(1, flt.Equal, 0.2)(inst))
        self.assertFalse(flt(1, flt.Equal, 0.3)(inst))

        self.assertTrue(flt(1, flt.NotEqual, 0.3)(inst))
        self.assertFalse(flt(1, flt.NotEqual, 0.2)(inst))

        self.assertTrue(flt(1, flt.Less, 0.3)(inst))
        self.assertFalse(flt(1, flt.Less, 0.2)(inst))

        self.assertTrue(flt(1, flt.LessEqual, 0.3)(inst))
        self.assertTrue(flt(1, flt.LessEqual, 0.2)(inst))
        self.assertFalse(flt(1, flt.LessEqual, 0.1)(inst))

        self.assertTrue(flt(1, flt.Greater, 0.1)(inst))
        self.assertFalse(flt(1, flt.Greater, 0.2)(inst))

        self.assertTrue(flt(1, flt.GreaterEqual, 0.1)(inst))
        self.assertTrue(flt(1, flt.GreaterEqual, 0.2)(inst))
        self.assertFalse(flt(1, flt.GreaterEqual, 0.3)(inst))

        self.assertTrue(flt(1, flt.Between, 0.05, 0.4)(inst))
        self.assertTrue(flt(1, flt.Between, 0.2, 0.4)(inst))
        self.assertTrue(flt(1, flt.Between, 0.05, 0.2)(inst))
        self.assertFalse(flt(1, flt.Between, 0.3, 0.4)(inst))

        self.assertFalse(flt(1, flt.Outside, 0.05, 0.4)(inst))
        self.assertFalse(flt(1, flt.Outside, 0.2, 0.4)(inst))
        self.assertFalse(flt(1, flt.Outside, 0.05, 0.2)(inst))
        self.assertTrue(flt(1, flt.Outside, 0.3, 0.4)(inst))

        self.assertTrue(flt(1, flt.IsDefined)(inst))
        self.assertFalse(flt(3, flt.IsDefined)(inst))

        self.assertRaises(ValueError, flt(1, -1, 1), inst)

    def test_position(self):
        inst = self.inst
        flt = FilterContinuous
        self.assertFalse(flt(0, flt.Equal, 0.2)(inst))
        self.assertTrue(flt(1, flt.Equal, 0.2)(inst))
        self.assertFalse(flt(2, flt.Equal, 0.2)(inst))
        self.assertFalse(flt(3, flt.Equal, 0.2)(inst))

        self.assertFalse(flt("a", flt.Equal, 0.2)(inst))
        self.assertTrue(flt("b", flt.Equal, 0.2)(inst))
        self.assertFalse(flt("c", flt.Equal, 0.2)(inst))
        self.assertFalse(flt("d", flt.Equal, 0.2)(inst))

        a, b, c, d = self.domain.attributes
        self.assertFalse(flt(a, flt.Equal, 0.2)(inst))
        self.assertTrue(flt(b, flt.Equal, 0.2)(inst))
        self.assertFalse(flt(c, flt.Equal, 0.2)(inst))
        self.assertFalse(flt(d, flt.Equal, 0.2)(inst))

    def test_nan(self):
        inst = self.inst
        flt = FilterContinuous

        self.assertFalse(flt(3, flt.Equal, 0.3)(inst))
        self.assertFalse(flt(3, flt.NotEqual, 0.3)(inst))
        self.assertFalse(flt(3, flt.Less, 0.2)(inst))
        self.assertFalse(flt(3, flt.LessEqual, 0.1)(inst))
        self.assertFalse(flt(3, flt.Greater, 0.2)(inst))
        self.assertFalse(flt(3, flt.GreaterEqual, 0.1)(inst))
        self.assertFalse(flt(3, flt.Between, 0.05, 0.4)(inst))
        self.assertFalse(flt(3, flt.Outside, 0.05, 0.4)(inst))

        self.assertTrue(flt(3, flt.Equal, np.nan)(inst))
        self.assertFalse(flt(3, flt.NotEqual, np.nan)(inst))

    def test_str(self):
        flt = FilterContinuous(1, FilterContinuous.Equal, 1)

        self.assertEqual(str(flt), "feature(1) = 1")

        flt = FilterContinuous("foo", FilterContinuous.Equal, 1)
        self.assertEqual(str(flt), "foo = 1")

        flt = FilterContinuous(self.domain[0], FilterContinuous.Equal, 1, 2)
        self.assertEqual(str(flt), "a = 1")

        flt.oper = flt.NotEqual
        self.assertEqual(str(flt), "a ≠ 1")

        flt.oper = flt.Less
        self.assertEqual(str(flt), "a < 1")

        flt.oper = flt.LessEqual
        self.assertEqual(str(flt), "a ≤ 1")

        flt.oper = flt.Greater
        self.assertEqual(str(flt), "a > 1")

        flt.oper = flt.GreaterEqual
        self.assertEqual(str(flt), "a ≥ 1")

        flt.oper = flt.Between
        self.assertEqual(str(flt), "1 ≤ a ≤ 2")

        flt.oper = flt.Outside
        self.assertEqual(str(flt), "not 1 ≤ a ≤ 2")

        flt.oper = flt.IsDefined
        self.assertEqual(str(flt), "a is defined")

        flt.oper = -1
        self.assertEqual(str(flt), "invalid operator")

    def test_eq(self):
        flt1 = FilterContinuous(1, FilterContinuous.Between, 1, 2)
        flt2 = FilterContinuous(1, FilterContinuous.Between, 1, 2)
        flt3 = FilterContinuous(1, FilterContinuous.Between, 1, 3)
        self.assertEqual(flt1, flt2)
        self.assertNotEqual(flt1, flt3)
        self.assertEqual(flt1.__dict__ == flt2.__dict__, flt1 == flt2)
        self.assertEqual(flt1.__dict__ == flt3.__dict__, flt1 == flt3)


class TestFilterDiscrete(unittest.TestCase):
    def test_eq(self):
        flt1 = FilterDiscrete(1, None)
        flt2 = FilterDiscrete(1, None)
        flt3 = FilterDiscrete(2, None)
        self.assertEqual(flt1, flt2)
        self.assertEqual(flt1.__dict__ == flt2.__dict__, flt1 == flt2)
        self.assertNotEqual(flt1, flt3)
        self.assertEqual(flt1.__dict__ == flt3.__dict__, flt1 == flt3)


class TestFilterString(unittest.TestCase):

    def setUp(self):
        self.data = Table("zoo")
        self.inst = self.data[0]  # aardvark

    def test_case_sensitive(self):
        flt = FilterString("name", FilterString.Equal, "Aardvark", case_sensitive=True)
        self.assertFalse(flt(self.inst))
        flt = FilterString("name", FilterString.Equal, "Aardvark", case_sensitive=False)
        self.assertTrue(flt(self.inst))

    def test_operators(self):
        flt = FilterString("name", FilterString.Equal, "aardvark")
        self.assertTrue(flt(self.inst))
        flt = FilterString("name", FilterString.Equal, "bass")
        self.assertFalse(flt(self.inst))

        flt = FilterString("name", FilterString.NotEqual, "bass")
        self.assertTrue(flt(self.inst))
        flt = FilterString("name", FilterString.NotEqual, "aardvark")
        self.assertFalse(flt(self.inst))

        flt = FilterString("name", FilterString.Less, "bass")
        self.assertTrue(flt(self.inst))
        flt = FilterString("name", FilterString.Less, "aa")
        self.assertFalse(flt(self.inst))

        flt = FilterString("name", FilterString.LessEqual, "bass")
        self.assertTrue(flt(self.inst))
        flt = FilterString("name", FilterString.LessEqual, "aardvark")
        self.assertTrue(flt(self.inst))
        flt = FilterString("name", FilterString.LessEqual, "aa")
        self.assertFalse(flt(self.inst))

        flt = FilterString("name", FilterString.Greater, "aa")
        self.assertTrue(flt(self.inst))
        flt = FilterString("name", FilterString.Greater, "aardvark")
        self.assertFalse(flt(self.inst))

        flt = FilterString("name", FilterString.GreaterEqual, "aa")
        self.assertTrue(flt(self.inst))
        flt = FilterString("name", FilterString.GreaterEqual, "aardvark")
        self.assertTrue(flt(self.inst))
        flt = FilterString("name", FilterString.GreaterEqual, "bass")
        self.assertFalse(flt(self.inst))

        flt = FilterString("name", FilterString.Between, "aa", "aardvark")
        self.assertTrue(flt(self.inst))
        flt = FilterString("name", FilterString.Between, "a", "aa")
        self.assertFalse(flt(self.inst))

        flt = FilterString("name", FilterString.Outside, "aaz", "bass")
        self.assertTrue(flt(self.inst))
        flt = FilterString("name", FilterString.Outside, "aardvark", "bass")
        self.assertFalse(flt(self.inst))

        flt = FilterString("name", FilterString.Contains, "ard")
        self.assertTrue(flt(self.inst))
        flt = FilterString("name", FilterString.Contains, "ra")
        self.assertFalse(flt(self.inst))

        flt = FilterString("name", FilterString.StartsWith, "aar")
        self.assertTrue(flt(self.inst))
        flt = FilterString("name", FilterString.StartsWith, "ard")
        self.assertFalse(flt(self.inst))

        flt = FilterString("name", FilterString.EndsWith, "aardvark")
        self.assertTrue(flt(self.inst))
        flt = FilterString("name", FilterString.EndsWith, "aard")
        self.assertFalse(flt(self.inst))

        flt = FilterString("name", FilterString.IsDefined)
        self.assertTrue(flt(self.inst))
        for s in ["?", "nan"]:
            self.inst["name"] = s
            flt = FilterString("name", FilterString.IsDefined)
            self.assertTrue(flt(self.inst))
        self.inst["name"] = ""
        flt = FilterString("name", FilterString.IsDefined)
        self.assertFalse(flt(self.inst))


class TestSameValueFilter(unittest.TestCase):
    def setUp(self):
        self.table = Table('zoo')

        self.attr_disc = self.table.domain["type"]
        self.attr_cont = self.table.domain["legs"]
        self.attr_meta = self.table.domain["name"]

        self.value_cont = 4
        self.value_disc = self.attr_disc.to_val("mammal")
        self.value_meta = self.attr_meta.to_val("girl")

    def test_same_value_filter_table(self):

        test_pairs = ((self.attr_cont, 4, self.value_cont),
                      (self.attr_disc, "mammal", self.value_disc),
                      (self.attr_meta, "girl", self.value_meta),)

        for var_index, value, num_value in test_pairs:
            filter_ = SameValue(var_index, value)(self.table)
            self.assertTrue(all(inst[var_index] == num_value for inst in filter_))

            filter_inverse = SameValue(var_index, value, negate=True)(self.table)
            self.assertTrue(all(inst[var_index] != num_value for inst in filter_inverse))

            self.assertEqual(len(filter_) + len(filter_inverse), len(self.table))


        for t1, t2 in itertools.combinations(test_pairs, 2):
            pos1, val1, r1 = t1
            pos2, val2, r2 = t2

            filter_1 = SameValue(pos1, val1)(self.table)
            filter_2 = SameValue(pos2, val2)(self.table)

            filter_12 = SameValue(pos2, val2)(SameValue(pos1, val1)(self.table))
            filter_21 = SameValue(pos1, val1)(SameValue(pos2, val2)(self.table))

            self.assertEqual(len(filter_21), len(filter_12))

            self.assertTrue(len(filter_1) >= len(filter_12))
            self.assertTrue(len(filter_2) >= len(filter_12))

            self.assertTrue(all(inst[pos1] == r1 and
                                inst[pos2] == r2 and
                                inst in filter_21
                                for inst in filter_12))
            self.assertTrue(all(inst[pos1] == r1 and
                                inst[pos2] == r2 and
                                inst in filter_12
                                for inst in filter_21))

    def test_same_value_filter_instance(self):
        inst = self.table[0]

        filter_ = SameValue(self.attr_disc, self.value_disc)(inst)
        self.assertEqual(filter_, inst[self.attr_disc] == self.value_disc)

        filter_n = SameValue(self.attr_disc, self.value_disc, negate=True)(inst)
        self.assertEqual(filter_n, inst[self.attr_disc] != self.value_disc)

    @patch('Orange.data.Table._filter_same_value', NIMOCK)
    def test_has_class_filter_not_implemented(self):
        self.test_same_value_filter_table()


class TestFilterReprs(unittest.TestCase):
    def setUp(self):
        self.table = Table('zoo')
        self.attr_disc = self.table.domain["type"]
        self.value_disc = self.attr_disc.to_val("mammal")
        self.vs = self.table.domain.variables

        self.table2 = Table("zoo")
        self.inst = self.table2[0]  # aardvark

    def test_reprs(self):
        flid = IsDefined(negate=True)
        flhc = HasClass()
        flr = Random()
        fld = FilterDiscrete(self.attr_disc, None)
        flsv = SameValue(self.attr_disc, self.value_disc, negate=True)
        flc = FilterContinuous(self.vs[0], FilterContinuous.Less, 5)
        flc2 = FilterContinuous(self.vs[1], FilterContinuous.Greater, 3)
        flv = Values([flc, flc2], conjunction=False, negate=True)
        flvf = ValueFilter(self.attr_disc)
        fls = FilterString("name", FilterString.Equal, "Aardvark", case_sensitive=False)
        flsl = FilterStringList("name", ["Aardvark"], case_sensitive=False)
        flrx = FilterRegex("name", "^c...$")

        filters = [flid, flhc, flr, fld, flsv, flc, flv, flvf, fls, flsl, flrx]

        for f in filters:
            repr_str = repr(f)
            new_f = eval(repr_str)
            self.assertEqual(repr(new_f), repr_str)



"""Pandas DataFrame↔Table conversion helpers"""
import numpy as np
import pandas as pd
from pandas.api.types import (
    is_categorical_dtype, is_object_dtype,
    is_datetime64_any_dtype, is_numeric_dtype,
)

from Orange.data import (
    Table, Domain, DiscreteVariable, StringVariable, TimeVariable,
    ContinuousVariable,
)

__all__ = ['table_from_frame', 'table_to_frame']


def table_from_frame(df, *, force_nominal=False):
    """
    Convert pandas.DataFrame to Orange.data.Table

    Parameters
    ----------
    df : pandas.DataFrame
    force_nominal : boolean
        If True, interpret ALL string columns as nominal (DiscreteVariable).

    Returns
    -------
    Table
    """

    def _is_discrete(s):
        return (is_categorical_dtype(s) or
                is_object_dtype(s) and (force_nominal or
                                        s.nunique() < s.size**.666))

    def _is_datetime(s):
        if is_datetime64_any_dtype(s):
            return True
        try:
            if is_object_dtype(s):
                pd.to_datetime(s, infer_datetime_format=True)
                return True
        except Exception:  # pylint: disable=broad-except
            pass
        return False

    # If df index is not a simple RangeIndex (or similar), put it into data
    if not (df.index.is_integer() and (df.index.is_monotonic_increasing or
                                       df.index.is_monotonic_decreasing)):
        df = df.reset_index()

    attrs, metas = [], []
    X, M = [], []

    # Iter over columns
    for name, s in df.items():
        name = str(name)
        if _is_discrete(s):
            discrete = s.astype('category').cat
            attrs.append(DiscreteVariable(name, discrete.categories.astype(str).tolist()))
            X.append(discrete.codes.replace(-1, np.nan).values)
        elif _is_datetime(s):
            tvar = TimeVariable(name)
            attrs.append(tvar)
            s = pd.to_datetime(s, infer_datetime_format=True)
            X.append(s.astype('str').replace('NaT', np.nan).map(tvar.parse).values)
        elif is_numeric_dtype(s):
            attrs.append(ContinuousVariable(name))
            X.append(s.values)
        else:
            metas.append(StringVariable(name))
            M.append(s.values.astype(object))

    return Table.from_numpy(Domain(attrs, None, metas),
                            np.column_stack(X) if X else np.empty((df.shape[0], 0)),
                            None,
                            np.column_stack(M) if M else None)


def table_to_frame(tab):
    """
    Convert Orange.data.Table to pandas.DataFrame

    Parameters
    ----------
    tab : Table

    Returns
    -------
    pandas.DataFrame
    """
    def _column_to_series(col, vals):
        result = ()
        if col.is_discrete:
            codes = pd.Series(vals).fillna(-1).astype(int)
            result = (col.name, pd.Categorical.from_codes(codes=codes, categories=col.values,
                                                          ordered=col.ordered))
        elif col.is_time:
            result = (col.name, pd.to_datetime(vals, unit='s').to_series().reset_index()[0])
        elif col.is_continuous:
            dt = float
            # np.nan are not compatible with int column
            nan_values_in_column = [t for t in vals if np.isnan(t)]
            if col.number_of_decimals == 0 and len(nan_values_in_column) == 0:
                dt = int
            result = (col.name, pd.Series(vals).astype(dt))
        elif col.is_string:
            result = (col.name, pd.Series(vals))
        return result

    def _columns_to_series(cols, vals):
        return [_column_to_series(col, vals[:, i]) for i, col in enumerate(cols)]

    x, y, metas = [], [], []
    domain = tab.domain
    if domain.attributes:
        x = _columns_to_series(domain.attributes, tab.X)
    if domain.class_vars:
        y_values = tab.Y.reshape(tab.Y.shape[0], len(domain.class_vars))
        y = _columns_to_series(domain.class_vars, y_values)
    if domain.metas:
        metas = _columns_to_series(domain.metas, tab.metas)
    all_series = dict(x + y + metas)
    original_column_order = [var.name for var in tab.domain.variables]
    unsorted_columns_df = pd.DataFrame(all_series)
    return unsorted_columns_df[original_column_order]



"""Common useful `QGraphicsView` classes that can be composed to achieve
desired functionality."""
from itertools import repeat

import numpy as np

from AnyQt.QtWidgets import QGraphicsView
from AnyQt.QtGui import QTransform
from AnyQt.QtCore import Qt


class ZoomableGraphicsView(QGraphicsView):
    """Zoomable graphics view.

    Composable graphics view that adds zoom functionality.

    It also handles automatic resizing of content whenever the window is
    resized.

    Right click will reset the zoom to a factor where the entire scene is
    visible.

    Parameters
    ----------
    scene : QGraphicsScene
    padding : int or tuple, optional
        Specify the padding around the drawn widgets. Can be an int, or tuple,
        the tuple can contain either 2 or 4 elements.

    Notes
    -----
    .. note:: This view will NOT consume the wheel event, so it would be wise
        to use this component in conjuction with the `PreventDefaultWheelEvent`
        in most cases.
    .. note:: This view does however consume the right mouse click event.

    """

    def __init__(self, scene, padding=(0, 0), **kwargs):
        self.zoom = 1
        self.scale_factor = 1 / 16
        # zoomout limit prevents the zoom factor to become negative, which
        # results in the canvas being flipped over the x axis
        self.__zoomout_limit_reached = False
        # Does the view need to recalculate the initial scale factor
        self.__needs_to_recalculate_initial = True
        self.__initial_zoom = -1

        self.__central_widget = None
        self.__set_padding(padding)

        super().__init__(scene, **kwargs)

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self.__needs_to_recalculate_initial = True

    def wheelEvent(self, event):
        self.__handle_zoom(event.angleDelta().y())
        super().wheelEvent(event)

    def mousePressEvent(self, event):
        # right click resets the zoom factor
        if event.button() == Qt.RightButton:
            self.reset_zoom()
        super().mousePressEvent(event)

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Plus:
            self.__handle_zoom(1)
        elif event.key() == Qt.Key_Minus:
            self.__handle_zoom(-1)

        super().keyPressEvent(event)

    def __set_padding(self, padding):
        # Allow for multiple formats of padding for convenience
        if isinstance(padding, int):
            padding = tuple(repeat(padding, 4))
        elif isinstance(padding, list) or isinstance(padding, tuple):
            if len(padding) == 2:
                padding = tuple(padding * 2)
        else:
            padding = 0, 0, 0, 0

        left, top, right, bottom = padding
        self.__padding = -left, -top, right, bottom

    def __handle_zoom(self, direction):
        """Handle zoom event, direction is positive if zooming in, otherwise
        negative."""
        if self.__zooming_in(direction):
            self.__reset_zoomout_limit()
        if self.__zoomout_limit_reached and self.__zooming_out(direction):
            return

        self.zoom += np.sign(direction) * self.scale_factor
        if self.zoom <= 0:
            self.__zoomout_limit_reached = True
            self.zoom += self.scale_factor
        else:
            self.setTransformationAnchor(self.AnchorUnderMouse)
            self.setTransform(QTransform().scale(self.zoom, self.zoom))

    @staticmethod
    def __zooming_out(direction):
        return direction < 0

    def __zooming_in(self, event):
        return not self.__zooming_out(event)

    def __reset_zoomout_limit(self):
        self.__zoomout_limit_reached = False

    def set_central_widget(self, widget):
        """Set the central widget in the view.

        This means that the initial zoom will fit the central widget, and may
        cut out any other widgets.

        Parameters
        ----------
        widget : QGraphicsWidget

        """
        self.__central_widget = widget

    def central_widget_rect(self):
        """Get the bounding box of the central widget.

        If a central widget and padding are set, this method calculates the
        rect containing both of them. This is useful because if the padding was
        added directly onto the widget, the padding would be rescaled as well.

        If the central widget is not set, return the scene rect instead.

        Returns
        -------
        QtCore.QRectF

        """
        if self.__central_widget is None:
            return self.scene().itemsBoundingRect().adjusted(*self.__padding)
        return self.__central_widget.boundingRect().adjusted(*self.__padding)

    def recalculate_and_fit(self):
        """Recalculate the optimal zoom and fits the content into view.

        Should be called if the scene contents change, so that the optimal zoom
        can be recalculated.

        Returns
        -------

        """
        if self.__central_widget is not None:
            self.fitInView(self.central_widget_rect(), Qt.KeepAspectRatio)
        else:
            self.fitInView(self.scene().sceneRect(), Qt.KeepAspectRatio)
        self.__initial_zoom = self.transform().m11()
        self.zoom = self.__initial_zoom

    def reset_zoom(self):
        """Reset the zoom to the optimal factor."""
        self.zoom = self.__initial_zoom
        self.__zoomout_limit_reached = False

        if self.__needs_to_recalculate_initial:
            self.recalculate_and_fit()
        else:
            self.setTransform(QTransform().scale(self.zoom, self.zoom))


class PannableGraphicsView(QGraphicsView):
    """Pannable graphics view.

    Enables panning the graphics view.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setDragMode(QGraphicsView.ScrollHandDrag)

    def enterEvent(self, event):
        self.viewport().setCursor(Qt.ArrowCursor)
        super().enterEvent(event)

    def mouseReleaseEvent(self, event):
        super().mouseReleaseEvent(event)
        self.viewport().setCursor(Qt.ArrowCursor)


class PreventDefaultWheelEvent(QGraphicsView):
    """Prevent the default wheel event.

    The default wheel event pans the view around, if using the
    `ZoomableGraphicsView`, this will prevent that behaviour.

    """

    def wheelEvent(self, event):
        event.accept()
